package at.gv.bmi.szr.client;

import java.util.NoSuchElementException;

/**
 * String Tokenizer zum Einlesen von CSV Dateien.
 * <p>
 * Im Unterschied zum normalen StringTokenizer werden bei
 * dieser Klasse auch Tokens zwischen unmittelbar 
 * aufeinanderfolgenden Trennzeichen sowie ein leeres Token nach
 * dem letzten Trennzeichen zurckgeliefert.
 * 
 * @author Gnter Pindhofer
 */
public class CsvStringTokenizer {

	private String toParse;
	private char delimiter;
	private int position;

	/**
	 * Erzeugt den Tokenizer.
	 * 
	 * @param toParse String der in Tokens zerlegt werden soll
	 * @param delimiter Trennzeichen
	 */
	public CsvStringTokenizer(String toParse, char delimiter) {
		super();

		this.toParse = toParse;
		this.delimiter = delimiter;
		this.position = 0;
	}

	/**
	 * Prft ob noch mehr Tokens vorhanden sind.
	 * 
	 * @return true wenn noch mindestens ein Token vorhanden ist
	 */
	public boolean hasMoreTokens() {
		return position <= toParse.length();
	}

	public boolean hasMore() {
		return hasMoreTokens();
	}

	/**
	 * Liefert das nchste Token.
	 * 
	 * @return das nchste Token aus dem Eingabestring
	 * @throws NoSuchElementException wenn kein weiteres
	 * Token vorhanden ist
	 */
	public String nextToken() throws NoSuchElementException {
		if (position > toParse.length())
			throw new NoSuchElementException();

		StringBuffer token = new StringBuffer();

		while ((position < toParse.length()) && (toParse.charAt(position) != delimiter)) {
			token.append(toParse.charAt(position));
			position++;
		}
		position++;
		return token.toString();
	}

	public String next() throws NoSuchElementException {
		return nextToken();
	}

	/**
	 * Liefert die Anzahl der Tokens.
	 * <p>
	 * Die Anzahl der Tokens ist immer um 1 grer als 
	 * die Anzahl der Trennzeichen.
	 *  
	 * @return die Anzahl der Tokens
	 */
	public int countTokens() {
		int tokens = 1;
		for (int i = 0; i < toParse.length(); i++)
			if (toParse.charAt(i) == delimiter)
				tokens++;
		return tokens;
	}
}
