/*
 * Created on Jul 5, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package at.gv.bmi.szr.client;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

/**
 * Erweiter den Hashtable soweit, dass mehrere Objekte unter einem Schlssel
 * gespeichert und gelesen werden knnen.
 * 
 * @author Oliver Ehrenmller
 */
public class HashVector extends Hashtable {

	private static final long serialVersionUID = 582464173176751938L;

	/**
	 * Um die eindeutige Identitt eines eigenen Vectors als Objekt
	 * festzustellen, werden mehrere Objekte unter einem Schlssel
	 * mit diesem Vector gespeichert. (Sonst knnen Daten mit 
	 * <code>java.util.Vector</code> verwechselt werden):
	 *
	 * @author Oliver Ehrenmller
	 */
	private class HVector extends Vector {

		private static final long serialVersionUID = -2340875875774665344L;

	}

	/**
	 * Wenn nach einem Objekt gesucht wird, mssen smtliche Listen
	 * geprft werden. 
	 * @see java.util.Map#containsValue(java.lang.Object)
	 */
	public boolean containsValue(Object object) {
		Enumeration enumeration = super.keys();
		while (enumeration.hasMoreElements()) {
			Object obj = enumeration.nextElement();
			if (obj instanceof HVector && ((HVector) obj).contains(object))
				return true;
		}
		return false;
	}

	public synchronized Enumeration elements() {
		return ((Vector) this.values()).elements();
	}

	public synchronized Object get(Object key) {
		Object object = super.get(key);
		return object instanceof HVector ? ((HVector) object).firstElement() : object;
	}

	public synchronized Vector getAll(Object key) {
		Object object = super.get(key);
		if (object == null)
			return new Vector();
		if (object instanceof HVector)
			return (HVector) object;
		Vector vector = new Vector();
		vector.add(object);
		return vector;
	}

	public synchronized Object put(Object[] keys, Object value) {
		for (int i = 0; i < keys.length; i++)
			put(keys[i], value);
		return null;
	}

	public synchronized Object put(Object key, Object value) {
		return this.put(key, value, false);
	}

	public synchronized Object put(Object key, Object value, boolean override) {
		if (super.containsKey(key)) {
			Object object = super.get(key);
			if (object instanceof HVector) {
				if (override)
					 ((HVector) object).clear();
				((HVector) object).add(value);
			} else {
				if (override)
					super.put(key, value);
				else {
					HVector hvector = new HVector();
					hvector.add(object);
					hvector.add(value);
					super.remove(key);
					super.put(key, hvector);
				}
			}
		} else
			super.put(key, value);
		return null;
	}

	public synchronized void putAll(Map map) {
		Iterator iterator = map.keySet().iterator();
		while (iterator.hasNext()) {
			Object key = iterator.next();
			this.put(key, map.get(key));
		}
	}

	public synchronized Object removeValue(Object value) {
		Iterator iterator = super.values().iterator();
		while (iterator.hasNext()) {
			Object object = iterator.next();
			if (object.equals(value)) {
				iterator.remove();
				return object;
			}
			if (object instanceof HVector) {
				HVector vector = (HVector) object;
				if (vector.remove(value)) {
					if (vector.size() == 0)
						iterator.remove();
					return value;
				}
			}
		}
		return null;
	}

	public int size() {
		int size = 0;
		Enumeration enumeration = super.keys();
		while (enumeration.hasMoreElements()) {
			Object object = enumeration.nextElement();
			if (object instanceof HVector)
				size += ((HVector) object).size();
			else
				size++;
		}
		return size;
	}

	public int size(String key) {
		if (!super.containsKey(key))
			return 0;
		Object object = super.get(key);
		return object instanceof HVector ? ((HVector) object).size() : 1;
	}

	public Collection values() {
		Vector vector = new Vector();
		Iterator iterator = super.values().iterator();
		while (iterator.hasNext()) {
			Object object = iterator.next();
			if (object instanceof HVector)
				vector.addAll((HVector) object);
			else
				vector.add(object);
		}
		return vector;
	}
}