package at.gv.bmi.szr.client;

import java.io.File;
import java.io.FileInputStream;

import org.apache.axis.client.Call;
import org.apache.axis.message.SOAPHeaderElement;
import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import at.gv.bmi.portal.pvp.PVP;
import at.gv.bmi.portal.util.PVPUtils;

public class PVPHeader {

	private float pvpVersion;
	private File pvpFile;

	public PVPHeader(float pvpVersion, File file) {
		this.pvpFile = file;
		this.pvpVersion = pvpVersion;
	}

	public void addToCall(Call call) throws Exception {
		if (this.pvpVersion == (float) 1.7)
			addPvp17Header(call);
		else if (this.pvpVersion == (float) 1.8)
			addPvp18Header(call);
	}

	private void addPvp17Header(Call call) throws Exception {
		if (this.pvpFile == null || !this.pvpFile.canRead())
			throw new Exception("Es wurde keine Property-Datei fr PVP-Infos gefunden bzw. gesetzt.");
		FileInputStream fis = new FileInputStream(this.pvpFile);
		PVP pvp = PVP.fromStream(fis);
		if (pvp.getX_AUTHENTICATE_cn() == null || pvp.getX_AUTHENTICATE_cn().equals(""))
			throw new Exception("PVP-Informationen in Datei '" + this.pvpFile.getCanonicalPath() + "' setzen!");
		Element pvpElement = PVPUtils.createHeaderElement(pvp);
		NodeList nl = pvpElement.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			SOAPHeaderElement sh = new SOAPHeaderElement((Element) nl.item(i));
			sh.setPrefix("pvp");
			sh.setNamespaceURI(PVP.PVP_NAMESPACE);
			call.addHeader(sh);
		}
	}

	private void addPvp18Header(Call call) throws Exception {
		DOMParser parser = new DOMParser();
		FileInputStream fileInputStream  = new FileInputStream(this.pvpFile);
		InputSource inputSource = new InputSource(fileInputStream);
		parser.parse(inputSource);
		Document document = parser.getDocument();
		SOAPHeaderElement element = new SOAPHeaderElement(document.getDocumentElement());
		call.addHeader(element);
	}

}
