package at.gv.bmi.szr.client;

import iaik.security.provider.IAIK;
import iaik.security.rsa.RSAPrivateKey;
import iaik.x509.SimpleChainVerifier;
import iaik.x509.X509Certificate;
import iaik.xml.crypto.XSecProvider;
import iaik.xml.crypto.dom.DOMMarshalContext;
import iaik.xml.crypto.dsig.ManifestImpl;
import iaik.xml.crypto.dsig.XMLSignatureFactory;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.security.KeyStore;
import java.security.Provider;
import java.security.Security;
import java.util.Iterator;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

import javax.crypto.Cipher;
import javax.xml.crypto.dsig.Manifest;
import javax.xml.crypto.dsig.Reference;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.dom.DOMValidateContext;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.rpc.ServiceException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.axis.AxisFault;
import org.apache.log4j.Logger;
import org.apache.log4j.xml.DOMConfigurator;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import at.gv.bmi.szr.BasiszahlZuBPKReturnType;
import at.gv.bmi.szr.FremdBPKRequestType;
import at.gv.bmi.szr.FremdBPKType;
import at.gv.bmi.szr.IdentityLinkType;
import at.gv.bmi.szr.SZR;
import at.gv.bmi.szr.SZRServiceLocator;
import at.gv.bmi.szr.SZRSoapBindingStub;
import at.gv.bmi.szr.ZMRAnwendungsIntegrationReturnType;
import at.gv.bmi.szr.key.KeyValueType;
import at.gv.bmi.szr.key.RSAKeyValueType;
import at.gv.bmi.szr.person.DeliveryAddressType;
import at.gv.bmi.szr.person.IdentificationType;
import at.gv.bmi.szr.person.PostalAddressType;

/**
 * SZR Axis Client, Hauptprogramm.
 * <p>
 * Test Programm zum Aufrufen diverser Funktionen des SZR
 * WebServices.
 * 
 * @author Gottfried Luef
 * @author Gnter Pindhofer
 * @author Oliver Ehrenmller
 */
public class SZRAxisClient {

	/**
	 * Startet den SZR Client.
	 * <p>
	 * siehe readme.txt fr Aufruf-Parameter
	 * 
	 * @param args Parameter aus der Command Line
	 * @throws Exception wenn ein Fehler auftritt
	 */
	public static void main(String[] args) {

		DOMConfigurator.configure("log4j.config");

		if (args.length == 0) {
			try {
				BufferedReader reader = new BufferedReader(new FileReader("readme.txt"));
				StringBuffer sb = new StringBuffer();
				while (reader.ready())
					sb.append(reader.readLine());
				logger.info(sb);
			} catch (IOException e) {
				logger.info("Parameter: -bpk | -transform | -sign | -verify | -wbpk | -basiszahl | -zmrai | -stammzahl | -test");
			}
			return;
		}

		logger.info("============================================================");

		try {
			Provider s = (Provider) Class.forName("com.sun.net.ssl.internal.ssl.Provider").newInstance();
			Security.addProvider(s);
			System.setProperty("java.protocol.handler.pkgs", "com.sun.net.ssl.internal.www.protocol");
		} catch (Exception e) {
			logger.error("An exception occured during initialization of the security environment", e);
			return;
		}

		Security.addProvider(new IAIK());
		Security.addProvider(new XSecProvider());

		SZRAxisClient client = new SZRAxisClient(args);
		client.urlToUse = client.params.getProperty("u", client.urlToUse);
		String keyStore = client.params.getProperty("ks");
		String keyStorePw = client.params.getProperty("kp", "changeit");
		String trustStore = client.params.getProperty("ts");
		boolean produktiv = client.urlToUse != null && (client.urlToUse.startsWith("szr.prod.") || client.urlToUse.startsWith("sz2.prod."));
		produktiv = Boolean.valueOf(client.params.getProperty("prod", String.valueOf(produktiv))).booleanValue();

		if (keyStore != null) {
			System.setProperty("javax.net.ssl.keyStore", keyStore);
			System.setProperty("javax.net.ssl.keyStorePassword", keyStorePw);
		}
		if (trustStore != null)
			System.setProperty("javax.net.ssl.trustStore", trustStore);

		if (!args[0].equals("-verify")) {
			try {
				logger.info("URL: " + client.urlToUse);
				float pvpVersion = Float.parseFloat(client.params.getProperty("pvpver", "1.8"));
				File file = new File(client.params.getProperty("pvp", "pvp.1_8.xml"));
				boolean checkversion = Boolean.valueOf(client.params.getProperty("version", "true")).booleanValue();
				client.initConnection(client.urlToUse, pvpVersion, file, checkversion);
			} catch (Exception e) {
				logger.error("\n\nAxisClient konnte nicht initialisiert werden:", e);
				return;
			}
		}

		try {
			TestCasePool pool = new TestCasePool();
			String dbFile = args[0].equals("-test") ? "standard-test.xml" : client.params.getProperty("d");
			if (dbFile == null) {
				pool.loadDefaults();
				logger.info("Testflle: default");
			} else {
				pool.load(dbFile);
				logger.info("Testflle aus folgender Datei gelesen: " + dbFile);
			}
			if (args[0].equals("-test")) {
				client.iterateTestCases(pool, TestCasePool.OPERATION_BPKZUBASISZAHL, produktiv);
				client.iterateTestCases(pool, TestCasePool.OPERATION_BASISZAHLZUBPK, produktiv);
				client.iterateTestCases(pool, TestCasePool.OPERATION_GETBPK, produktiv);
				client.iterateTestCases(pool, TestCasePool.OPERATION_GETIDENTITYLINK, produktiv);
				client.iterateTestCases(pool, TestCasePool.OPERATION_TRANSFORMBPK, produktiv);
			} else if (args[0].equals("-bpk"))
				client.iterateTestCases(pool, TestCasePool.OPERATION_GETBPK, produktiv);
			else if (args[0].equals("-stammzahl"))
				client.iterateTestCases(pool, TestCasePool.OPERATION_GETSTAMMZAHL, produktiv);
			else if (args[0].equals("-transform"))
				client.iterateTestCases(pool, TestCasePool.OPERATION_TRANSFORMBPK, produktiv);
			else if (args[0].equals("-sign"))
				client.iterateTestCases(pool, TestCasePool.OPERATION_GETIDENTITYLINK, produktiv);
			else if (args[0].equals("-basiszahl"))
				client.iterateTestCases(pool, TestCasePool.OPERATION_BASISZAHLZUBPK, produktiv);
			else if (args[0].equals("-verify")) {
				if (args.length < 2)
					logger.error("Please specify the name of the file to be verified");
				else
					for (int i = 1; i < args.length; i++) {
						if (!args[i].startsWith("-")) {
							client.processVerify(args[i]);
						}
					}
			} else if (args[0].equals("-wbpk")) {
				TestCase testCase = pool.getTestCase(TestCasePool.OPERATION_BPKZUBASISZAHL, 0);
				try {
					client.processBPKzuBasiszahl(testCase, produktiv);
				} catch (AxisFault fault) {
					logger.error(new TestCaseResult(testCase, fault).getResultText());
				}
			} else if (args[0].equals("-basiszahl")) {
				client.iterateTestCases(pool, TestCasePool.OPERATION_BASISZAHLZUBPK, produktiv);
			} else if (args[0].equals("-zmrai")) {
				client.iterateTestCases(pool, TestCasePool.OPERATION_ZMRINTEGRATION, produktiv);
			} else
				logger.info("Please specify argument (-bpk | -transform | -sign | -verify <fn> | -wbpk | -zmrai | -test)");
		} catch (Exception e) {
			logger.error("UNEXCEPTED CLIENT EXCEPTION", e);
		}
		logger.info("============================================================");
	}

	private SZRServiceLocator locator;
	private final String[] paramKeys = new String[] { "bereich", "p", "z", "tb", "d", "u", "ks", "kp", "ts", "pvp", "pvpver", "version" };
	private PropertiesVector params;
	private SZR stub;

	private String validationCertPath = null;

	private static final String VERSION = "1.3";

	private static final String PERSONRECORD_NAMESPACE = "http://reference.e-government.gv.at/namespace/persondata/20020228#";

	private String urlToUse = null;

	private static Logger logger = Logger.getLogger(SZRAxisClient.class);

	public SZRAxisClient() {
		super();
		this.params = new PropertiesVector();
	}
	/**
	 * Konstruktor.
	 * <p>
	 * Initialisiert Klassenfelder.
	 *
	 * @param args Commandline-Parameter vom Aufruf der Klasse.
	 * @throws MalformedURLException	Tritt auf, sollte die Serveraddresse ungltig sein.
	 * @throws ServiceException		Tritt auf, sollte zum Server keine Verbindung 
	 * 						hergestellt werden knnen.
	 */
	public SZRAxisClient(String[] args) {
		this();
		parseParams(args);
	}

	/**
	 * Prft und Initialisiert die URL des SZR-Servers und setzt den Pfad fr die PVP-Informationen.
	 * @param url URL zum SZR-Server bzw Portal.
	 * @param pvpVersion Definiert die zu verwendende Version des PVP (zb. 1.7).
	 * @param pvpProperties Pfad zu Properties-Datei binhaltend die PVP-Infos fr den Header.
	 * @throws MalformedURLException Geworfen, sollte die bergebene Url nicht konform sein.
	 * @throws ServiceException Geworfen, sollte ein Fehler bei der Client-Initialisierung auftreten.
	 */
	public void initConnection(String url, float pvpVersion, File pvpProperties) throws MalformedURLException, ServiceException {
		initConnection(url, pvpVersion, pvpProperties, true);
	}

	/**
	 * Prft und Initialisiert die URL des SZR-Servers und setzt den Pfad fr die PVP-Informationen.
	 * @param url URL zum SZR-Server bzw Portal.
	 * @param pvpVersion Definiert die zu verwendende Version des PVP (zb. 1.7).
	 * @param pvpProperties Pfad zu Properties-Datei binhaltend die PVP-Infos fr den Header.
	 * @param validateVersion <code>true</code>, wenn die getVersion-Operation ausgefhrt werden soll
	 * 						  und die Version von Client und Server verglichen wird.
	 * @throws MalformedURLException Geworfen, sollte die bergebene Url nicht konform sein.
	 * @throws ServiceException Geworfen, sollte ein Fehler bei der Client-Initialisierung auftreten.
	 */
	public void initConnection(String url, float pvpVersion, File pvpProperties, boolean validateVersion) throws MalformedURLException, ServiceException {
		url = solveURL(url);
		if (this.urlToUse != null && this.urlToUse.equalsIgnoreCase(url) && this.locator != null)
			return;
		this.locator = new SZRServiceLocator();
		this.stub = this.locator.getSZR(new URL(url));
		logger.info("PVP-Version: " + pvpVersion);
		if (pvpProperties != null)
			 ((SZRSoapBindingStub) this.stub).setPVPHeader(new PVPHeader(pvpVersion, pvpProperties));
		this.urlToUse = url;

		if (validateVersion) {
			try {
				logger.info("Prfe Version... ");
				String serverVersion = this.stub.getVersion(null);
				logger.info("Client=" + SZRAxisClient.VERSION + ", Server=" + serverVersion);
				if (!serverVersion.equals(SZRAxisClient.VERSION)) {
					if (isServerOld(serverVersion, SZRAxisClient.VERSION)) {
						throw new ServiceException("Version des Server ist lter als Client (Server=" + serverVersion + ", Client=" + SZRAxisClient.VERSION + ")");
					} else
						logger.warn("Warnung: Client lter als Server, Anfragen fr neuen Client.");
				}
			} catch (RemoteException e) {
				throw new ServiceException("Fehler bei Versionsabfrage: " + e.getMessage());
			}
		}
	}

	/**
	 * Vergleicht zwei Versions-Strings (zb 1.3.23) und gibt an, ob Server zu alt ist.
	 * Liefert <code>true</code>, wenn die Server-Version lter als die des Clients ist. 
	 * Liefert <code>false</code>, wenn Version des Client gleich oder neuer ist. 
	 * @param serverVersion Version-String des Servers (zb. 1.3.43)
	 * @param clientVersion Version-String des Clients (zb. 1.3.43)
	 * @return <code>true</code>, falls Server lter als Client ist.
	 */
	private boolean isServerOld(String serverVersion, String clientVersion) {
		StringTokenizer server = new StringTokenizer(serverVersion, ".");
		StringTokenizer client = new StringTokenizer(clientVersion, ".");
		while (server.hasMoreTokens()) {
			if (!client.hasMoreTokens())
				return false;
			int s = Integer.parseInt(server.nextToken());
			int c = Integer.parseInt(client.nextToken());
			if (s < c)
				return true;
			if (c < s)
				return false;
		}
		return true;
	}

	/**
	 * Lst eine URL auf, die durch Symbolnamen angegeben wurden. 
	 * @param url Symbolname einer URL
	 * @return URL
	 * @throws MalformedURLException Sollte die URL einen unzulssigen Wert haben.
	 */
	private String solveURL(String url) throws MalformedURLException {
		if (url.startsWith("szr") || url.startsWith("sz2")) {
			ResourceBundle bundle = ResourceBundle.getBundle("at/gv/bmi/szr/client/url");
			url = bundle.getString(url);
			if (url == null || url.trim().length() == 0)
				throw new MalformedURLException("Url '" + url + "' konnte nicht gefunden werden");
			logger.info("URL solved to: " + url);
		}
		return url;
	}

	/**
	 * Erzeugt ein Array von <code>TransformTargetType</code>. Das wird fr TransformBPK
	 * bentigt. Diese TargetTypes stellen die Informationen zu den Bereichen
	 * in die transformiert werden soll. Die Daten werden dazu aus den Attributen des
	 * Testfalles gelesen. 
	 * @param testCase	Testfall aus dem die Informationen fr die Targets gelesen werden.
	 * @return	Array von TransformTargets fr transformBPK.
	 */
	private FremdBPKRequestType[] generateFremdBPKRequests(TestCase testCase) {
		String[] tBereich = this.params.containsKey("tb") ? this.params.getAllProperties("tb") : testCase.getAllProperties(TestCasePool.D.BPK_TARGET);
		RSAKeyValueType rsa_type = testCase.getFirstRSAKey();
		if (rsa_type == null) {
			RSAPrivateKey key = TestCasePool.sampleRSA();
			rsa_type = new RSAKeyValueType();
			rsa_type.setModulus(TestCasePool.toBase64(key.getModulus().toByteArray()));
			rsa_type.setExponent(TestCasePool.toBase64(key.getPublicExponent().toByteArray()));
		}
		FremdBPKRequestType[] targets = new FremdBPKRequestType[tBereich.length];
		for (int i = 0; i < tBereich.length; i++) {
			targets[i] = new FremdBPKRequestType();
			// targets[i].setVKZ("L9-AL");
			targets[i].setVKZ("BMI");
			targets[i].setBereichsKennung(tBereich[i]);
		}
		return targets;
	}

	/**
	 * Liest aus einem <code>TestCasePool</code> alle Testflle fr die
	 * angegebene Operation. Dann werden Attribute gelesen und fr jeden Testfall
	 * die entsprechende Operation ausgefhrt.
	 * @param pool	Testdatensammlung aus der die Daten gelesen werden.
	 * @throws RemoteException Tritt auf, falls bei der Verbindung ein Fehler auftritt.
	 */
	public void iterateTestCases(TestCasePool pool, String operation, boolean produktiv) {
		Iterator it = pool.getTestCases(operation);
		while (it.hasNext()) {
			TestCase testCase = (TestCase) it.next();
			try {
				processSingleTestCase(testCase, operation, produktiv);
			} catch (RemoteException exception) {
				logger.info(new TestCaseResult(testCase, exception).getValidateText());
			}
		}
	}

	/**
	 * Fhrt fr einen TestCase die Operation durch. Zustzlich noch die Angabe
	 * ob ein Produktiv-System getestet wird (Fr Datenverifikation)
	 * @param testCase 	Testfall
	 * @param operation	Operation (zb. BPK)
	 * @param produktiv <code>true</code>, falls Produktionsdaten verwendet werden sollen.
	 */
	public void processSingleTestCase(TestCase testCase, String operation, boolean produktiv) throws RemoteException {
		String probeBPK = testCase.properties.getProperty(produktiv ? TestCasePool.D.BPK_PRODUKTION : TestCasePool.D.BPK_TEST);
		if (probeBPK != null)
			testCase.properties.setProperty(TestCasePool.D.BPK, probeBPK, true);
		try {
			IdentificationType id = testCase.personInfo.getPerson().getIdentification();
			if (id.getType() != null && id.getType().equals("urn:publicid:gv.at:cdid+TEST")&& testCase.properties.containsKey(TestCasePool.D.BPK))
				id.setValue(testCase.properties.getProperty(TestCasePool.D.BPK));
		} catch (NullPointerException npe) {
		}
		if (operation.equals(TestCasePool.OPERATION_TRANSFORMBPK))
			processTransformBPK(testCase);
		else if (operation.equals(TestCasePool.OPERATION_GETBPK)) {
			String testBPK = testCase.properties.getProperty(produktiv ? TestCasePool.D.BPK_PRODUKTION : TestCasePool.D.BPK_TEST);
			String bpk = processGetBPK(testCase);
			if (testBPK != null && testBPK.trim().length() > 0 && !bpk.equals(testBPK))
				throw new RemoteException("Erhaltene BPK '" + bpk + "' entspricht nicht der erwarteten: " + testBPK);
		}else if (operation.equals(TestCasePool.OPERATION_GETSTAMMZAHL)) {
			String testSZ = testCase.properties.getProperty(produktiv ? TestCasePool.D.SZ_PROD : TestCasePool.D.SZ_TEST);
			String sz = processGetStammzahl(testCase);
			if (testSZ != null && testSZ.trim().length() > 0 && !sz.equals(testSZ))
				throw new RemoteException("Erhaltene Stammzahl '" + sz + "' entspricht nicht der erwarteten: " + testSZ);
		} else if (operation.equals(TestCasePool.OPERATION_GETIDENTITYLINK))
			processGetIdentityLink(testCase, produktiv);
		else if (operation.equals(TestCasePool.OPERATION_BPKZUBASISZAHL))
			processBPKzuBasiszahl(testCase, produktiv);
		else if (operation.equals(TestCasePool.OPERATION_BASISZAHLZUBPK)) {
			String testBPK = testCase.properties.getProperty(produktiv ? TestCasePool.D.BPK_PRODUKTION : TestCasePool.D.BPK_TEST);
			String bpk = processBasiszahlZuBPK(testCase, produktiv);
			if (testBPK != null && testBPK.trim().length() > 0 && !bpk.equals(testBPK))
				throw new RemoteException("Erhaltene BPK '" + bpk + "' entspricht nicht der erwarteten: " + testBPK);
		} else if (operation.equals(TestCasePool.OPERATION_ZMRINTEGRATION))
			processZMRAnwendungsintegration(testCase, produktiv);
		else
			throw new RemoteException("Operation '" + operation + "' unbekannt");
		logger.info(new TestCaseResult(testCase, null).getValidateText());
	}

	/**
	 * Vergleicht im bergebenen String-Array ob Prefixe fr Parameter vorkommen.
	 * Ist das der Fall werden die Anhnge an dem Prefix in einem Property-Table unter
	 * dem Schlssel des Prefixes gespeichert.
	 *
	 * @param args Commandline-Parameter als String-Array
	 */
	public void parseParams(String[] args) {
		int level = 0, index = 0;
		for (int i = 0; i < args.length; i++) {
			for (int j = 0; j < this.paramKeys.length; j++)
				if (args[i].startsWith("-" + this.paramKeys[j]) && level < this.paramKeys[j].length() && args[i].length() > this.paramKeys[j].length()) {
					level = this.paramKeys[j].length();
					index = j;
				}
			if (index > 0)
				this.params.setProperty(this.paramKeys[index], args[i].substring(this.paramKeys[index].length() + 1));
			level = index = 0;
		}
	}

	/**
	 * Durchfhrt eine BPKzuBasiszahl-Anfrage an einen SZR-Server. Dabei
	 * werden die Parameter der Funktion fr die Informationen der Anfrage
	 * genutzt.
	 * @param zmr Basiszahlen mit '|' getrennt
	 * @param bereich Bereich
	 * @param bpk Bereichsspezifische Personenkennung
	 * @throws RemoteException
	 */
	public String processBPKzuBasiszahl(String zmr, String bereich, String bpk) throws RemoteException {
		StringTokenizer tok = new StringTokenizer(zmr, "|");
		String[] zmra = new String[tok.countTokens()];
		int zmri = 0;
		while (tok.hasMoreTokens()) {
			zmra[zmri] = tok.nextToken();
			zmri++;
		}
		return processBPKzuBasiszahl(zmra, bereich, bpk);
	}

	/**
	 * Durchfhrt eine BPKzuBasiszahl-Anfrage an einen SZR-Server. Dabei
	 * werden die Parameter der Funktion fr die Informationen der Anfrage
	 * genutzt.
	 * @param basiszahlen Basiszahlen
	 * @param bereich Bereich
	 * @param bpk Bereichsspezifische Personenkennung
	 * @throws RemoteException
	 */
	public String processBPKzuBasiszahl(String[] basiszahlen, String bereich, String bpk) throws RemoteException {
		logger.info("Basiszahl zu BPK wird ermittelt ...");
		String bz = this.stub.BPKzuBasiszahl(bereich, bpk, basiszahlen);
		logger.info("Basiszahl is: " + bz);
		return bz;
	}

	/**
	 * Durchfhrt eine ZMRAnwendungsintegration-Anfrage an einen SZR-Server. Dabei
	 * werden die Parameter der Funktion fr die Informationen der Anfrage
	 * genutzt.
	 * @param bereich Bereich
	 * @param targets FremdBPK-Bereiche + VKZ
	 * @param fremdbPKs BMI/ZP-Fremd-bPKs
	 * @throws RemoteException
	 */
	public void processZMRAnwendungsintegration(TestCase testCase, boolean produktiv) throws RemoteException {
		logger.info("ZMR-Anwendungsintegration wird durchgefhrt ...");
		String bereich = testCase.getProperty(TestCasePool.D.BPK_BEREICH);
		FremdBPKRequestType[] targets = generateFremdBPKRequests(testCase);
		String[] fremdbPKs = testCase.getAllProperties(TestCasePool.D.ZMR_FREMDBPK);
		logger.info("Anfrage-Bereich: " + bereich);
		for (int i = 0; i < fremdbPKs.length; i++) 
			logger.info("Anfrage-ZMR-FremdbPK: " + fremdbPKs[i]);
		ZMRAnwendungsIntegrationReturnType[] result = this.stub.ZMRAnwendungsIntegration(bereich, targets, fremdbPKs);
		for (int i = 0; i < result.length; i++) {
			logger.info("     bPK [" + i + "]: " + result[i].getBPK());
			FremdBPKType[] fremdbpks = result[i].getFremdBPKs();
			for (int j = 0; j < fremdbpks.length; j++)
				logger.info("fremdbPK [" + j + "] " + fremdbpks[j].getBereichsKennung() + ": "+ fremdbpks[j].getFremdBPK());
		}
	}

	/**
	 * Durchfhrt eine BPKzuBasiszahl-Anfrage an einen SZR-Server. Dabei
	 * werden die Inhalte des Testfalles fr die Informationen der Anfrage
	 * genutzt.
	 * @param testCase Beinhaltet Testdaten
	 * @return Basiszahl
	 * @throws RemoteException
	 */
	public String processBasiszahlZuBPK(TestCase testCase, boolean produktiv) throws RemoteException {
		String bereich = testCase.getProperty(TestCasePool.D.BPK_BEREICH);
		String basiszahl = testCase.getProperty(TestCasePool.D.BPK_BASISZAHL);
		FremdBPKRequestType[] frequests = generateFremdBPKRequests(testCase);
		logger.info("BPK ber BasiszahlZuBPK wird ermittelt ...");
		BasiszahlZuBPKReturnType result = this.stub.basiszahlZuBPK(new String[] { basiszahl }, bereich, frequests)[0];
		logger.info("BPK is: " + result.getBPK());
		FremdBPKType[] fbpks = result.getFremdBPKs();
		if (fbpks != null)
			for (int i = 0; i < fbpks.length; i++) {
				logger.info("FremdBPK fr " + fbpks[i].getBereichsKennung() + " = " + fbpks[i].getFremdBPK());
				try {
					Cipher ci = Cipher.getInstance("RSA/ECB/OAEP", "IAIK");
					RSAPrivateKey key = getPrivateKey(fbpks[i].getBereichsKennung());
					ci.init(Cipher.DECRYPT_MODE, key);
					byte[] plain = ci.doFinal(TestCasePool.toByteArray(fbpks[i].getFremdBPK()));
					logger.info("decrypted: " + new String(plain));
				} catch (Exception e) {
					logger.error("Fehler bei entschlsseln der FremdBPK", e);
				}
			}
		return result.getBPK();
	}

	/**
	 * Durchfhrt eine BPKzuBasiszahl-Anfrage an einen SZR-Server. Dabei
	 * werden die Inhalte des Testfalles fr die Informationen der Anfrage
	 * genutzt.
	 * @param testCase Beinhaltet Testdaten
	 * @return Basiszahl
	 * @throws RemoteException
	 */
	public String processBPKzuBasiszahl(TestCase testCase, boolean produktiv) throws RemoteException {
		String bereich = testCase.getProperty(TestCasePool.D.BPK_BEREICH);
		String bpk = testCase.getProperty(TestCasePool.D.BPK);
		if (bpk == null)
			bpk = testCase.getProperty(produktiv ? TestCasePool.D.BPK_PRODUKTION : TestCasePool.D.BPK_TEST);
		String[] basiszahlen = testCase.getAllProperties(TestCasePool.D.BPK_BASISZAHL);
		return processBPKzuBasiszahl(basiszahlen, bereich, bpk);
	}

	/**
	 * Durchfhrt eine GetBPK-Anfrage an einem SZR-Server.
	 * @param testCase Testfall, aus dem die Daten fr die Anfrage gelesen werden.
	 * @throws RemoteException
	 */
	public String processGetBPK(TestCase testCase) throws RemoteException {
		logger.info("************************************************************");
		logger.info("Testfall: " + testCase.getName());
		String bereich = this.params.getProperty("bereich", testCase.getProperty(TestCasePool.D.BPK_BEREICH));
		try {
			IdentificationType id = testCase.personInfo.getPerson().getIdentification();
			if (id.getType() != null && id.getValue() == null)
				id.setValue(testCase.getProperty(TestCasePool.D.BPK));
		} catch (NullPointerException npe) {
		}
		logger.info("BPK fr " + testCase + " in Bereich " + bereich + " wird ermittelt ...");
		String bpk = this.stub.getBPK(testCase.personInfo, bereich);
		logger.info("BPK (" + bereich + ") is: " + bpk);
		String compareBPK = testCase.getProperty(TestCasePool.D.BPK);
		if (compareBPK != null && !bpk.equals(compareBPK))
			throw new RemoteException("Erhaltene BPK '" + bpk + "' von erwarteter '" + compareBPK + "' verschieden.");
		return bpk;
	}

	/**
	 * Durchfhrt eine GetStammzahl-Anfrage an einem SZR-Server.
	 * @param testCase Testfall, aus dem die Daten fr die Anfrage gelesen werden.
	 * @throws RemoteException
	 */
	public String processGetStammzahl(TestCase testCase) throws RemoteException {
		logger.info("************************************************************");
		logger.info("Testfall: " + testCase.getName());
		logger.info("Stammzahl fr " + testCase + " wird ermittelt ...");
		String stammzahl = this.stub.getStammzahl(testCase.personInfo);
		logger.info("Stammzahl ist: " + stammzahl);
		String compareSZ = testCase.getProperty(TestCasePool.D.SZ);
		if (compareSZ != null && !stammzahl.equals(compareSZ))
			throw new RemoteException("Erhaltene Stammzahl '" + stammzahl + "' von erwarteter '" + compareSZ + "' verschieden.");
		return stammzahl;
	}

	/**
	 * Durchfhrt eine GetIdentityLink-Anfrage an einem SZR-Server.
	 * @param testCase Testfall, aus dem die Daten fr die Anfrage gelesen werden.
	 * @throws RemoteException
	 */
	public void processGetIdentityLink(TestCase testCase, boolean produktiv) throws RemoteException {
		KeyValueType[] keys = testCase.getKeyArray();
		logger.info("************************************************************");
		logger.info("Testfall: " + testCase.getName());
		logger.info("Personenbindung fr " + testCase + " wird ermittelt ...");
		IdentityLinkType idLink = this.stub.getIdentityLink(testCase.personInfo, keys);
		Element resultEle = idLink.getAssertion();
		String filePath = this.params.getProperty("p", testCase.getProperty("filepath"));
		if (filePath != null) {
			if (filePath.charAt(filePath.length() - 1) != File.separatorChar)
				filePath += File.separatorChar;
			try {
				Transformer t = TransformerFactory.newInstance().newTransformer();
				File file = new File(filePath + testCase.personInfo.getPerson().getName().getFamilyName() + ".xml");
				FileOutputStream out = new FileOutputStream(file);
				t.transform(new DOMSource(resultEle), new StreamResult(out));
				out.close();
			} catch (Exception e) {
				logger.error(e);
			}
		}
		if (idLink.getPersonInfo().getRegularDomicile() != null) {
			PostalAddressType padr = idLink.getPersonInfo().getRegularDomicile();
			logger.info("Ort: " + padr.getMunicipality());
			logger.info("PLZ: " + padr.getPostalCode());
			if (padr.getDeliveryAddress() != null) {
				DeliveryAddressType dadr = padr.getDeliveryAddress();
				 logger.info("Strae: "+ dadr.getStreetName().toString());
				 if (dadr.getBuildingNumber() != null)
					 logger.info("Hausnummer: "+dadr.getBuildingNumber().toString());
				 if (dadr.getUnit() != null)
					 logger.info("Nutzungseinheitnummer: "+dadr.getUnit().toString());
			}
		}
		
		logger.info("Erzeugte Personenbindung wird validiert ...");
		String SZ_PROP = produktiv ? TestCasePool.D.SZ_PROD : TestCasePool.D.SZ_TEST;
		String sz = testCase.getProperty(SZ_PROP);
		boolean valid = verifyPersbin(resultEle.getOwnerDocument(), produktiv, produktiv, sz);
		logger.info("Validation result: " + (valid ? "VALID" : "INVALID"));
		if (!valid)
			throw new RemoteException("Signatur konnte nicht verifiziert werden");
	}

	public RSAPrivateKey getPrivateKey(String bereich) throws Exception {
		KeyStore ks = KeyStore.getInstance("JKS");
		String path = (this.validationCertPath == null || this.validationCertPath.trim().length() == 0 ? "keys" : this.validationCertPath) + "/fremdbpks.jks";
		ks.load(new FileInputStream(path), "changeit".toCharArray());
		java.security.interfaces.RSAPrivateKey key = (java.security.interfaces.RSAPrivateKey) ks.getKey(bereich.substring(bereich.length() - 2), "changeit".toCharArray());
		if (key == null)
			throw new NullPointerException("Kein Schlssel fr " + bereich + " zum Entschlsseln vorhanden!");
		RSAPrivateKey prkey = new RSAPrivateKey(key.getEncoded());
		return prkey;
	}

	/**
	 * Durchfhrt eine TransformBPK-Anfrage an einem SZR-Server.
	 * @param testCase Testfall, aus dem die Daten fr die Anfrage gelesen werden.
	 * @throws RemoteException
	 */
	public void processTransformBPK(TestCase testCase) throws RemoteException {
		logger.info("************************************************************");
		logger.info("Testfall: " + testCase.getName());
		String bereich = testCase.getProperty(TestCasePool.D.BPK_BEREICH);
		String bpk = testCase.getProperty(TestCasePool.D.BPK);
		if (bereich == null || bpk == null)
			throw new NullPointerException("Bereich/BPK fehlt in Identification-Tag");
		FremdBPKRequestType[] targets = generateFremdBPKRequests(testCase);
		int i = 0;
		StringBuffer sb = new StringBuffer();
		while (i < targets.length - 1)
			sb.append(targets[i].getVKZ() + "(" + targets[i++].getBereichsKennung() + "), ");
		sb.append(targets[i].getVKZ() + "(" + targets[i++].getBereichsKennung() + ")");
		logger.info("BPK '" + bpk + "' von " + testCase + " wird fr folgende Bereiche transformiert: " + sb);
		FremdBPKType[] results = this.stub.transformBPK(testCase.personInfo, bpk, bereich, "SZR-Axis-Client Standard-Test", targets);
		for (i = 0; i < results.length; i++) {
			logger.info("BPK fr " + results[i].getBereichsKennung() + " ist: " + results[i].getFremdBPK());
			try {
				Cipher ci = Cipher.getInstance("RSA/ECB/OAEP", "IAIK");
				RSAPrivateKey key = getPrivateKey(results[i].getBereichsKennung());
				ci.init(Cipher.DECRYPT_MODE, key);
				byte[] plain = ci.doFinal(TestCasePool.toByteArray(results[i].getFremdBPK()));
				logger.info("decrypted: " + new String(plain));
			} catch (Exception e) {
				logger.error("Fehler bei entschlsseln der FremdBPK", e);
			}
		}
	}

	/**
	 * Liest aus einer Datei ein XML, das eine Signatur beinhlt und versucht 
	 * diese zu validieren.
	 * @param file Signatur beinhaltende XML-Datei
	 */
	public void processVerify(String file) {
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setNamespaceAware(true);
			Document source = factory.newDocumentBuilder().parse(new File(file));
			logger.info("Verifying signature in file " + file);
			boolean valid = verifyPersbin(source, true, true, "RzLLN4RN6B/iMzKGqSds5A==");
			logger.info("Validation result: " + (valid ? "VALID" : "INVALID"));
		} catch (Exception e) {
			logger.error(e);
		}
	}

	/**
	 * Verifiziert ein DOM-Dokument auf eine gltige Personenbindung. Geprft 
	 * werden dabei folgende Punkte:
	 * <ul>
	 * 		<li>Stammzahlenvergleich</li>
	 * 		<li>Suchen und Auslesen des Zertifikats aus der Personenbindung</li>
	 * 		<li>Verifizieren der Signatur</li>
	 * 		<li>Verifizieren des Manifests</li>
	 * 		<li>Ausgabe des Zertifikats-Subjects</li>
	 * 		<li>Prfen der Zertifikatskette</li>
	 * </ul>
	 * Die Prfung der Zertifikatskette und der Vergleich der Stammzahl kann ber
	 * Parameter der Methode forciert werden.
	 * @param document	DOM-Dokument, dass die zu prfende Personenbindung beinhaltet
	 * @param denySelfsignedCertificates	<code>TRUE</code>, wenn das Zertifikat
	 * 										von A-Trust stammen soll.
	 * @param stammzahl	BASE64-Codierte Stammzahl, die in der
	 * 					Personenbindung erwartet wird. <code>null</code>,
	 * 					wenn keine Stammzahl-Prfung erfolgen soll.
	 * @return <code>TRUE</code>, wenn alle Prfungen erfolgt sind und die Personenbindung gltig ist.
	 * 			<code>FALSE</code> sonst.
	 */
	public boolean verifyPersbin(Document document, boolean denySelfsignedCertificates, boolean useProductiveCertifcates, String stammzahl) {
		try {
			boolean valid = true;
			logger.info("VERIFYING (IAIK)");

			if (stammzahl != null && stammzahl.trim().length() > 0) {
				NodeList list = document.getElementsByTagNameNS(PERSONRECORD_NAMESPACE, "Value");
				String sz;
				if (list.getLength() == 0 || (sz = list.item(0).getFirstChild().getNodeValue()) == null) {
					logger.error("Stammzahl nicht gefunden!");
					return false;
				}
				if (!stammzahl.equals(sz)) {
					logger.error("Erhaltene Stammzahl " + sz + " nicht erwartet, erwartet " + stammzahl);
					return false;
				}
				logger.info("Stammzahl geprft.");
			}

			XMLSignatureFactory factory = (XMLSignatureFactory) XMLSignatureFactory.getInstance("DOM");

			logger.info("Get X509Certificate");
			NodeList list = document.getElementsByTagNameNS(XMLSignature.XMLNS, "X509Certificate");
			if (list.getLength() == 0)
				throw new Exception("Cannot find X509Certificate");
			String certBase64 = list.item(0).getFirstChild().getNodeValue().trim();
			X509Certificate cert = new X509Certificate(TestCasePool.toByteArray(certBase64));
			logger.info("Found X509Certificate: " + cert.getSubjectDN());
			logger.info("Verifying Core-Signature ...");
			list = document.getElementsByTagNameNS(XMLSignature.XMLNS, "Signature");
			if (list.getLength() == 0)
				throw new Exception("Cannot find Signature element");
			DOMValidateContext valContext = new DOMValidateContext(cert.getPublicKey(), list.item(0));
			XMLSignature signature = factory.unmarshalXMLSignature(valContext);
			boolean coreValidity = signature.validate(valContext);
			logger.info("Signature " + (coreValidity == false ? "failed" : "passed") + " core validation.");
			valid &= coreValidity;
			boolean sv = signature.getSignatureValue().validate(valContext);
			valid &= sv;
			logger.info("signature validation status: " + sv);
			Iterator i = signature.getSignedInfo().getReferences().iterator();
			for (int j = 0; i.hasNext(); j++) {
				Reference reference = ((Reference) i.next());
				boolean refValid = reference.validate(valContext);
				valid &= refValid;
				logger.info("  ref[" + j + "] URI=" + reference.getURI() + ", validity status=" + refValid);
			}

			logger.info("Verifying Manifest ...");
			list = document.getElementsByTagNameNS(XMLSignature.XMLNS, "Manifest");
			if (list.getLength() == 0)
				throw new Exception("Cannot find Manifest element");
			DOMMarshalContext marContext = new DOMMarshalContext(valContext);
			Manifest manifest = new ManifestImpl(marContext, list.item(0));
			Iterator iterator = manifest.getReferences().iterator();
			for (int j = 0; iterator.hasNext(); j++) {
				Reference reference = (Reference) iterator.next();
				boolean refValid = reference.validate(valContext);
				valid &= refValid;
				logger.info("  ref[" + j + "] URI=" + reference.getURI() + ", validity status=" + refValid);
			}

			logger.info("Verifying Certificate ...");
			logger.info("Zertificate-Subject: " + cert.getSubjectDN());

			if (useProductiveCertifcates) {
				if (!"EMAIL=dsk@dsk.gv.at,serialNumber=700247499804,CN=Dr. Waltraut Kotschy,OU=Datenschutzkommission,O=Bundeskanzleramt,C=AT".equals(cert.getSubjectDN().getName())) {
					
					logger.error("Personenbindung wurde mit nicht erwarteten Zertifikat signiert!");
					return false;
				}
			} else {								
//				 Cert vor 10.1.2008 if (!"serialNumber=491162583653,CN=Dr. Waltraut Kotschy,OU=Datenschutzkommission,O=Bundeskanzleramt,C=AT".equals(cert.getSubjectDN().getName())) {
				if (!"EMAIL=dsk@dsk.gv.at,serialNumber=491162583653,CN=Dr. Waltraut Kotschy,OU=Datenschutzkommission,O=Bundeskanzleramt,C=AT".equals(cert.getSubjectDN().getName()) && !"CN=hw-szr-sw,OU=BMI,O=BMI,ST=AT".equals(cert.getSubjectDN().getName())) {
					logger.error("Personenbindung wurde mit nicht erwarteten Zertifikat signiert!");
					return false;
				}
			}

			logger.info("Zertificate-Issuer:  " + cert.getIssuerDN());
			if (cert.getSubjectDN().equals(cert.getIssuerDN())) {
				logger.info("Zertificate is selfsigned!");
				return !denySelfsignedCertificates;
			}
			SimpleChainVerifier cv = new SimpleChainVerifier();
			X509Certificate[] x509s = new X509Certificate[3];
			x509s[0] = cert;
			String cert1 = useProductiveCertifcates ? "a-sign-corporate-light-03.der" : "a-sign-corp-light-test-02.der";
			String cert2 = useProductiveCertifcates ? "A-Trust-nQual-03.der" : "A-Trust-Test-Qual_02.cacert.der";

			if (this.validationCertPath == null || this.validationCertPath.trim().length() == 0) {
				x509s[1] = new X509Certificate(new FileInputStream("keys/" + cert1));
				x509s[2] = new X509Certificate(new FileInputStream("keys/" + cert2));
			} else {
				x509s[1] = new X509Certificate(new FileInputStream(this.validationCertPath + cert1));
				x509s[2] = new X509Certificate(new FileInputStream(this.validationCertPath + cert2));
			}
			logger.info("Trusting Zertificate-Subject: " + x509s[2].getSubjectDN());
			cv.addTrustedCertificate(x509s[2]);
			try {
				cv.verifyChain(x509s);
				logger.info("Certificate-Chain passed validation.");
			} catch (Exception exception) {
				valid = false;
				logger.info("Certificate-Chain failed validation.");
			}
			return valid;
		} catch (Exception exception) {
			logger.error("Exception caught: " + exception.getClass() + ", Message=" + exception.getMessage());
			return false;
		}
	}

	/**
	 * Getter fr den Pfad unter dem die A-Trust-Zertifikate gefunden 
	 * werden sollen. Ntig fr die Validierung einer Personenbindung,
	 * wenn die Zertifikate nicht im regulren <i>keys</i>-Unterverzeichnis
	 * zu finden sind.
	 * @return Pfad zu den A-Trust-Zertifikaten
	 */
	public String getValidationCertPath() {
		return validationCertPath;
	}

	/**
	 * Setter fr den Pfad unter dem die A-Trust-Zertifikate gefunden 
	 * werden sollen. Ntig fr die Validierung einer Personenbindung,
	 * wenn die Zertifikate nicht im regulren <i>keys</i>-Unterverzeichnis
	 * zu finden sind.
	 * @param string	Pfad zu den A-Trust-Zertifikaten.
	 */
	public void setValidationCertPath(String string) {
		validationCertPath = string;
	}

}