/*
 * Created on Aug 16, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package at.gv.bmi.szr.client;

import java.util.Iterator;
import java.util.Vector;

import at.gv.bmi.szr.PersonInfoType;
import at.gv.bmi.szr.key.KeyValueType;
import at.gv.bmi.szr.key.RSAKeyValueType;

/**
 * Speichert einen gesamten Testfall. Ein Testfall ist einer Operation zugewiesen.
 * Ein Testfall beinhaltet die Daten fr die Anfrage (bzw. Operation) an einem SZR-Server.
 * Zustzlich knnen weitere Attribute gespeichert werden, die zb. operationsabhngig sind.
 * Weiters kann ein <code>faultCode</code> angegeben werden. Der besagt, das eine 
 * Fehlermeldung erwartet wird. Sollte einer Angegeben werden, wird die Antwort gegen
 * den gemeldeten Fehlercode validiert. Sollte die Antwort unerwartet ausfallen, ist
 * der Testfall nicht erfolgreich.
 *
 * @author Oliver Ehrenmller
 */
public class TestCase {
	/**
	 * Bezeichnung des Testfalles.
	 */
	private String name;
	/**
	 * Bezeichnung der Operation. Mgliche Werte sind <code>OPERATION_GETBPK</code>, 
	 * <code>OPERATION_GETIDENTITYLINK</code> und OPERATION_TRANSFORMBPK</code>.
	 */
	String[] operations;
	/**
	 * Sollte der Testfall eine Fehlermeldung testen, kann hier der erwartete
	 * faultCode stehen. zb. "F230" fr Person nicht gefunden.
	 */
	String faultCode;
	/**
	 * Schlsseln (RSA, DSA, ECDSA) die fr Anfragen bentigt werden.
	 */
	Vector key;
	/**
	 * Kommentare und Erluterungen zu diesem Testfall.
	 */
	String comment;
	/**
	 * Personeninformationen fr die Anfrage.
	 */
	PersonInfoType personInfo;
	/**
	 * Zustzliche Attribute.
	 */
	PropertiesVector properties;

	/**
	 * Konstruktor. Es muss eine Operation gesetzt werden. 
	 * @param name Bezeichnung des Testfalles.
	 */
	TestCase(String name) {
		this.name = name;
		this.personInfo = new PersonInfoType();
		this.properties = new PropertiesVector();
		this.operations = new String[0];
		this.key = new Vector();
	}

	/**
	 * Gibt die Verfgbaren Schlsseln zu diesem Testfall als Array zurck.
	 * @return	verfgbare Schlsseln fr Anfrage.
	 */
	public KeyValueType[] getKeyArray() {
		KeyValueType[] array = new KeyValueType[this.key.size()];
		for (int i = 0; i < array.length; i++)
			array[i] = (KeyValueType) this.key.get(i);
		return array;
	}

	/**
	 * Zusammenfassender Ausgabestring.
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		try {
			return this.name + " (" + this.personInfo.getPerson().getName().getGivenName()
				+ " "
				+ this.personInfo.getPerson().getName().getFamilyName()
				+ ", "
				+ this.personInfo.getPerson().getDateOfBirth()
				+ ")";
		} catch (NullPointerException e) {
			return personInfo.toString();
		}
	}

	/**
	 * Zugriff auf ein Attribut.
	 * @param name	Name des Attributes.
	 * @return	Wert des zugehrigen Attributes vom Testfall. Sollte
	 * 			in den Testfall-Attributen, das Attribut nicht
	 * 			vorhanden sein, wird in den Attributen der zugehrigen
	 * 			Operation nachgesehen. <code>null</code>, falls kein Attribut
	 * 			gefunden wird. 
	 */
	public String getProperty(String name) {
		return getProperty(name, null);
	}

	/**
	 * Zugriff auf ein Attribut.
	 * @param name Name des Attributes.
	 * @param defaultValue	Rckgabewert. falls kein Attribut gefunden wird. 
	 * @return	Wert des zugehrigen Attributes vom Testfall. Gibt den Wert von Parameter
	 * 			<code>defaultValue</code> zurck, falls hier kein Attribut 
	 * 			gefunden wird.
	 */
	public String getProperty(String name, String defaultValue) {
		return this.properties.getProperty(name, defaultValue);
	}

	/**
	 * Es besteht, die Mglichkeit unter einem Namen mehrere Attribut-Werte zu speichern.
	 * Sollte das erfolgt sein, kann ber diese Methode, der Zugriff auf smtliche
	 * Attributwerte erfolgen.
	 * @param name	Name des Attributes.
	 * @return	Smtliche Werte die unter einem Attributnamen gefunden werden.
	 */
	public String[] getAllProperties(String name) {
		return this.properties.getAllProperties(name);
	}

	public RSAKeyValueType getFirstRSAKey() {
		Iterator iterator = this.key.iterator();
		while (iterator.hasNext()) {
			KeyValueType type = (KeyValueType) iterator.next();
			if (type.getRSAKeyValue() != null)
				return type.getRSAKeyValue();
		}
		return null;
	}
	/**
	 * Liefert den Namen des Testfalles.
	 * @return Bezeichnung des Testfalles.
	 */
	public String getName() {
		return name;
	}

}
