/*
 * Created on Aug 16, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package at.gv.bmi.szr.client;

import org.apache.axis.AxisFault;

/**
 * Wertet das Ergebnis eines durchgefhrten Testfalles aus. Vorwiegend
 * prft diese Klasse, ob von einem Testfall eine Exception erwartet
 * wird und vergleicht diese mit einer als Ergebnis geltenden.
 * 
 * @author Oliver Ehrenmller
 */
public class TestCaseResult {

	private TestCase testCase;
	private Exception exception;
	private StringBuffer validateText;
	private String resultText;
	private boolean failed = false;
	private boolean backendFailed = false;

	public TestCaseResult(TestCase testCase) {
		this(testCase, null);
	}

	public TestCaseResult(TestCase testCase, Exception exception) {
		this(testCase, exception, null);
	}

	public TestCaseResult(TestCase testCase, Exception exception, String text) {
		this.testCase = testCase;
		this.exception = exception;
		this.validateText = new StringBuffer(text == null ? "" : text);
		validateTestCase();
	}

	/**
	 * Bearbeitet das Ergebnis eines Testfalles und gibt gegebenenfalls die erhaltene
	 * Fehlermeldung dazu aus.
	 * <p>
	 * Es wird verglichen, ob die im Testfall angegebene Fehlermeldung sofern vorhanden,
	 * im AxisFault enthalten ist. Abhngig davon, wird ausgegeben, ob der Testfall
	 * erfolgreich war oder nicht.  
	 */
	private boolean validateTestCase() {
		this.resultText = null;
		if (this.exception != null) {
			if (this.exception instanceof AxisFault) {
				AxisFault fault = (AxisFault) this.exception;
				this.validateText.append("SOAP fault received");
				this.validateText.append("\n     Code:   " + fault.getFaultCode());
				this.validateText.append("\n     Actor:  " + fault.getFaultActor());
				this.validateText.append("\n     String: " + fault.getFaultString());
				if (testCase != null && testCase.faultCode != null) {
					this.validateText.append("\nSOAP fault expected: " + this.testCase.faultCode);
					if (!fault.getFaultCode().getLocalPart().equals(testCase.faultCode)) {
						this.resultText =
							"Different SOAP fault received: " + fault.getFaultCode() + ", " + fault.getFaultString();
						this.failed = true;
					}
				} else {
					this.resultText =
						"Not expected SOAP fault received: " + fault.getFaultCode() + ", " + fault.getFaultString();
					this.failed = true;
				}
				String faultCode = fault.getFaultCode().getLocalPart();
				if (this.failed && faultCode.equals("F599"))
					this.failed = false;
				this.backendFailed = faultCode.equals("F502") || faultCode.equals("F504");

			} else {
				this.validateText.append("Exception received");
				this.validateText.append("\n     Class:   " + this.exception.getClass());
				this.validateText.append("\n     Message: " + this.exception.getMessage());
				this.resultText = "Not expected Exception received: " + this.exception.getMessage();
				this.failed = true;
			}
		} else if (this.testCase != null) {
			this.validateText.append("NO SOAP fault received");
			if (this.testCase.faultCode != null) {
				this.validateText.append("\nSOAP fault expected: " + this.testCase.faultCode);
				this.resultText = "Expected SOAP fault not received";
				this.failed = true;
			}
			if (this.testCase.comment != null)
				this.validateText.append("\nKommentar: " + testCase.comment);
		}
		this.validateText.append("\nTestresult: ");
		this.validateText.append(getResultText());
		return !this.failed;
	}
	public boolean failed() {
		return this.failed;
	}

	public String getValidateText() {
		return this.validateText.toString();
	}

	public String getResultText() {
		StringBuffer sb = new StringBuffer();
		sb.append(this.failed ? "FAILED" : "PASSED");
		if (this.resultText != null) {
			sb.append(" (");
			sb.append(this.resultText.trim());
			sb.append(")");
		}
		return sb.toString();
	}

	public String toString() {
		return this.getResultText();
	}

	public boolean backendFailed() {
		return this.backendFailed;
	}
}
