/*
 * Copyright 2004 Bundeskanzleramt sterreich
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



package moazs.api; 

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;
import java.util.Random;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import moazs.Constants;
import moazs.modell.DeliveryQuality;
import moazs.modell.DualDelivery;
import moazs.modell.Identification;
import moazs.modell.Internetaddress;
import moazs.modell.Payload;
import moazs.modell.Postaladdress;
import moazs.modell.Rcpt;
import moazs.modell.Request;
import moazs.modell.Status;
import moazs.modell.Telephone;
import moazs.modell.XmlDocument;
import moazs.modell.Zuseserver;
import moazs.util.HibernateUtil;
import moazs.util.MoaConfig;
import moazs.util.MoaXMLUtils;
import moazs.util.MoaXPath;

import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.attachments.AttachmentPart;
import org.apache.axis.attachments.Attachments;
import org.apache.log4j.Logger;
import org.apache.xml.serialize.XMLSerializer;
import org.apache.xpath.XPathAPI;
import org.hibernate.HibernateException;
import org.hibernate.classic.Session;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import at.gv.egovernment.moa.util.DOMUtils;

/**
* Hilfsklasse mit der der DOMTree des Requests in die MoaZSDatenstruktur gemappt wird.
*/
public class MoaMapper {
//    private static String NS_MZS = null;
//    private static String NS_PERS = null;
    private static String NS_XSI;
    private static String BPK_TYPE;
    private static String BPK_TYPEV;
    private static MoaConfig mcfg = null;
   
    static private Logger logger = Logger.getLogger(MoaMapper.class);

    static {
        mcfg = new MoaConfig();
        
//        NS_MZS = mcfg.getProperty("mzs", "Namespaces");
//        NS_PERS= mcfg.getProperty("p1",	 "Namespaces");
        NS_XSI = mcfg.getProperty("xsi", "Namespaces");

        BPK_TYPE  = mcfg.getBereichskennung();
        BPK_TYPEV = mcfg.getBereichskennungEncrypted();
    }
    
   /**
    * Speichert den aktuellen Request und prft auf vollstndigkeit
     * @param sess Session
     * @param in Document
     * @return RequestObjekt
     * @throws net.sf.hibernate.HibernateException
     * @throws MoaException
     */
    public static Request saveRequest(Session sess, Document in) throws HibernateException, MoaException  {
        if (in == null) {
          return null;
        }

        Request r = new Request();
        Rcpt rcpt = saveRcpt(HibernateUtil.getCurrentSession(), in);
        r.setRcpt(rcpt);
        r.setRequiresEncryption(getRequiresEncryption( in));
        r.setZuseserver(getZuseServer(in, rcpt, r.getRequiresEncryption()));
        r.setSenderProfileId(getSenderProfileId(in));
        r.setSignatureProfileId(getSignatureProfileId(in));
        r.setAppDeliveryId(getAppDeliveryId(in));
        r.setPrivateDelivery(getPrivateDelivery(in, r.getSenderProfileId()));
        r.setDocumentClass(getDocumentClass(in));
        
        // check for dual delivery
        r.setDualDelivery(getDualDelivery(in));
        
        boolean isSync = getSync(in);
        if (isSync) {
          Random rand = new Random(System.currentTimeMillis());
          SimpleDateFormat format = new SimpleDateFormat("yyyyMMddHHmmSS");
          // create unique id
          String id =  format.format(new Date()) + String.valueOf(rand.nextInt(1000000));
          r.setSyncId(id);
        }
        try {
          r.setDeliveryQuality( (DeliveryQuality)(sess.find("from dq in class moazs.modell.DeliveryQuality where name='"+ getDeliveryQuality(in)+"'").get(0)));
        } catch (java.lang.IndexOutOfBoundsException iob) {
            throw new MoaException(Constants.ERROR_REQUEST_DELIVERY_QUALITY_UNKNOWN, "DeliveryQuality ("+getDeliveryQuality(in)+") unknown.");
        }

        // check for valid delivery quality
        boolean isPrivateDelivery = r.isPrivateDelivery();
        
        // delivery quality for private deliveries
        if (isPrivateDelivery) {
          if (!DeliveryQuality.PRIVATE_DQ.contains(r.getDeliveryQuality().getName())) {
            throw new MoaException(Constants.ERROR_REQUEST_DELIVERY_QUALITY_NOT_ALLOWED, "DeliveryQuality (" + r.getDeliveryQuality().getName() + ") not allowed for private delivery.");
          }
        } else {
          
          // check for document class
          if (r.getDocumentClass() != null) {
            throw new MoaException(Constants.ERROR_REQUEST_DOCUMENT_CLASS_NOT_ALLOWED, "The use of the document class element is only allowed for private deliveries.");
          }
          
          // check for dual delivery
          if ((r.getDualDelivery() == null) && DeliveryQuality.DUAL_DELIVERY_DQ.contains(r.getDeliveryQuality().getName())) {
            throw new MoaException(Constants.ERROR_REQUEST_DELIVERY_QUALITY_NOT_ALLOWED, "DeliveryQualities RSb and RSb+ only allowed for dual delivery.");
          } else if (DeliveryQuality.PRIVATE_DQ.contains(r.getDeliveryQuality().getName())) {
            throw new MoaException(Constants.ERROR_REQUEST_DELIVERY_QUALITY_NOT_ALLOWED, "Private delivery quality (" + r.getDeliveryQuality().getName() + ") not allowed for governmental delivery.");
          }
        }
        
        
        
        r.setStatus( (Status)sess.load(Status.class, new Long(0)));

        /* Vollstndigkeitsprfungen der Personendaten */
        if (rcpt.getZusebpk() == null) {
          if (rcpt.getPhysicalPerson() == true) {
            if (rcpt.getGivenname()  == null) {
              throw new MoaException("GivenName missing");
            }
            if (rcpt.getFamilyname() == null) {
              throw new MoaException("FamilyName missing");
            }
            if (r.getDeliveryQuality().getName().equals(DeliveryQuality.RSA) && rcpt.getDateofbirth() == null) {
              throw new MoaException("DateOfBirth missing (RSa requires date of birth).");
            }
          } else {
              if (rcpt.getFullname() == null) {
                throw new MoaException("Fullname missing");
              }
          }
          if (!hasBPK(rcpt) && !hasPostAddress(rcpt) &&!hasInternetAddress(rcpt)) {
            throw new MoaException("BPK or Address Missing");
          }
          /*if (hasBPK(rcpt) && rcpt.getDateofbirth() == null) {
            throw new MoaException("DateOfBirth missing");
          }*/
        }
        
        r.setTs(new Date()); 
        r.setRetryTs(new Date()); 

        sess.save(r);

        getPayload(HibernateUtil.getCurrentSession(), r, in);

        return r;
        
    }
    
    /**
    * extrahiert die EmpfaengerDaten und -Adressen
     * @param sess Session
     * @param in Document
     * @return EmpfaengerObject
     * @throws net.sf.hibernate.HibernateException
     * @throws MoaException
     */
    public static Rcpt saveRcpt(Session sess, Document in) throws HibernateException, MoaException {
        if (in == null ) {
          return null;
        }

        Rcpt rcpt = new Rcpt();
        Element physicalPerson = MoaXPath.getFirstElement( in.getDocumentElement(), "mzs:Receiver/p:PhysicalPerson");
        Element corporateBody = MoaXPath.getFirstElement( in.getDocumentElement(), "mzs:Receiver/p:CorporateBody");

        // Test auf Axis oder .net WSDL2irgendwas converter die die Person objekte mit 
        // xsi:type casten
        try {
            if ( physicalPerson == null && corporateBody == null ) {
                Element p = MoaXPath.getFirstElement( in.getDocumentElement(), "mzs:Receiver/p:Person");
                if (p.getAttributeNS(NS_XSI, "type").indexOf("PhysicalPersonType") != -1) {
                    physicalPerson = p;
                } else if ( p.getAttributeNS(NS_XSI, "type").indexOf("CorporateBodyType") != -1) {
                    corporateBody = p;
                }
            }
        } catch (Exception e) {
        }

        if (physicalPerson != null) {
          logger.debug("Found physical person.");
            rcpt.setPhysicalPerson(true);
            String givenName = MoaXPath.getFirstString( physicalPerson, "p:Name/p:GivenName/text()");
            if (givenName != null) {
              rcpt.setGivenname(givenName);
            }
            Element name = MoaXPath.getFirstElement( physicalPerson, "p:Name/p:FamilyName");
            if (name!=null) {
                NamedNodeMap atts = name.getAttributes();
                for(int i=0;i<atts.getLength(); i++) {
                    if (atts.item(i).getLocalName().equals("prefix")) {
                        rcpt.setPrefix(atts.item(i).getNodeValue());
                    } else if (atts.item(i).getLocalName().equals("primary")) {
                        rcpt.setPrimary( !atts.item(i).getNodeValue().equals("false") );
                    }
                }
                String familyName = MoaXPath.getFirstString( physicalPerson, "p:Name/p:FamilyName/text()");
                if (familyName != null) {
                  rcpt.setFamilyname(name.getFirstChild().getNodeValue());
                }
            }
            sess.save(rcpt);

            SimpleDateFormat fmt = new SimpleDateFormat("yyyy-MM-dd");
            String dateofbirth = MoaXPath.getFirstString( physicalPerson, "p:DateOfBirth/text()");
            if (dateofbirth != null) {
                try {
                    Date dob =  fmt.parse(dateofbirth);
                    if (new Date().before(dob)) {
                        throw new MoaException("DateOfBirth is in the Future ...");
                    }
                    rcpt.setDateofbirth ( dob);
                } catch (java.text.ParseException pe) {
                    throw new MoaException("Unparseable Date ...");

                }
           }
       } else if (corporateBody != null) {
            logger.debug("Corporate Body");
            rcpt.setPhysicalPerson(false);

            rcpt.setFullname( MoaXPath.getFirstString( corporateBody,      "p:FullName/text()"));
            rcpt.setOrganisation( MoaXPath.getFirstString( corporateBody , "p:Organisation/text()"));
        } else {
            throw new MoaException("Unknown PersonType"); 
        }

        sess.save(rcpt);
      
        boolean axis = false;
        if (MoaXPath.getFirstElement(in.getDocumentElement(), "mzs:Receiver/p:Address")!=null) {
            axis = true;
        }

        /* Telefon */
        Element[] telephone = null;
        if (!axis) {
            telephone = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver/p:TelephoneAddress");
        } else {
            telephone = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver/p:Address[contains(@xsi:type,'TelephoneAddressType')]");
        }

        for (int i =0; i<telephone.length; i++) {
            Telephone tel = new Telephone();
            tel.setRcpt(rcpt);
            tel.setValue( MoaXPath.getFirstString(telephone[i], "p:Number/text()"));
            if (tel.getValue()==null) throw new MoaException("Telephone number missing");
            sess.save(tel);
            rcpt.addTelephone(tel);
        }

        /* Email */
        Element[] internet;
        if (!axis) {
            internet = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver/p:InternetAddress");
        } else {
            internet = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver/p:Address[contains(@xsi:type,'InternetAddressType')]");
        }
        for (int i =0; i<internet.length; i++) {
            String e = MoaXPath.getFirstString(internet[i], "p:Address/text()");
            if (e != null)  {
                Internetaddress email = new Internetaddress();
                email.setRcpt(rcpt);
                email.setValue( e );
                sess.save(email);
                rcpt.addInternetaddress(email);
            }
        }
        
        /* PostAddresse */
        Element[] post;
        if (!axis) {
            post = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver/p:PostalAddress");
        } else {
            post = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver/p:Address[contains(@xsi:type,'PostalAddressType')]");
        }
        for (int i =0; i<post.length; i++) {
            logger.debug("Postaladdress ...");
            Postaladdress addr = new Postaladdress();
            addr.setRcpt(rcpt);
            addr.setCountrycode( MoaXPath.getFirstString(post[i], "p:CountryCode/text()"));
            if (addr.getCountrycode() != null && addr.getCountrycode().length() > 2) {
              throw new MoaException("Invalid CountryCode");
            }
            if (addr.getCountrycode() == null) {
              logger.debug("No country code set, setting default country code \"" + Constants.DEFAULT_COUNTRYCODE +"\"");
              addr.setCountrycode(Constants.DEFAULT_COUNTRYCODE);
            }
            addr.setPostalcode( MoaXPath.getFirstString(post[i], "p:PostalCode/text()"));
            if (addr.getPostalcode() == null) {
              throw new MoaException("PostalCode missing");
            }
            addr.setMunicipality( MoaXPath.getFirstString(post[i], "p:Municipality/text()"));
            if (addr.getMunicipality() == null) {
              throw new MoaException("Municipality missing");
            }
            addr.setStreetname(MoaXPath.getFirstString(post[i], "p:DeliveryAddress/p:StreetName/text()"));
            if (addr.getStreetname() == null) {
              throw new MoaException("StreetName missing");
            }
            addr.setBuilding( MoaXPath.getFirstString(post[i], "p:DeliveryAddress/p:BuildingNumber/text()"));
            if (addr.getBuilding() == null) {
              throw new MoaException("BuildingNumber missing");
            }
            addr.setUnit(MoaXPath.getFirstString(post[i], "p:DeliveryAddress/p:Unit/text()"));
            addr.setDoor(MoaXPath.getFirstString(post[i], "p:DeliveryAddress/p:DoorNumber/text()"));
            sess.save(addr);
            rcpt.addPostaladdress(addr);
        }

        /* Id nummern - bpks */
        Element[] ident = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Receiver//p:Identification");
        for (int i =0; i<ident.length; i++) {
            Identification id = new Identification();
            id.setRcpt(rcpt);
            id.setValue( MoaXPath.getFirstString( ident[i], "p:Value/text()"));
            if (id.getValue()==null || id.getValue().trim().equals("")) throw new MoaException("ID Value missing");
            id.setIdType( MoaXPath.getFirstString( ident[i], "p:Type/text()"));
            if (id.getIdType()==null || id.getIdType().trim().equals("")) throw new MoaException("ID Type missing");

            sess.save(id);
            rcpt.addIdentification(id);

            // natural persons
            if (rcpt.getPhysicalPerson()) {
              if (id.getIdType().trim().equals( BPK_TYPE ) || 
                  id.getIdType().trim().equals( BPK_TYPEV )) {
                if (id.getIdType().trim().equals( BPK_TYPE )) {
                  logger.debug("Found ZUSE-bPK for recipient.");
                } else {
                  logger.debug("Found encrypted ZUSE-bPK for recipient.");
                }
                rcpt.setZusebpk(id.getValue());
              }
            // non natural persons
            } else {
              // must be in list of corporate bodies
              if (mcfg.getCorporateBodies().contains(id.getIdType().trim())) {
                logger.debug("Found stammzahl for recipient.");
                rcpt.setZusebpk(id.getValue());
              }
            }
        }
        
        return rcpt;
    }
    
    /**
    * holt alle PayloadDaten
     * @param sess Session
     * @param r Request
     * @param in Document
     * @throws MoaException
     */
    public static void getPayload(Session sess, Request r, Document in) throws MoaException {
        Element xmldoc = MoaXPath.getFirstElement(in.getDocumentElement(), "mzs:XMLDocument");
        if (xmldoc!=null) {
            XmlDocument xmldocument = new XmlDocument();
 
            xmldocument.setXmlProfileId( MoaXPath.getFirstString( xmldoc, "mzs:XMLProfileID/text()"));
            
            if (xmldocument.getXmlProfileId() == null) {
                xmldocument.setMimetype( MoaXPath.getFirstString( xmldoc, "mzs:MIMEType/text()"));
                xmldocument.setResultingMimetype( MoaXPath.getFirstString( xmldoc, "mzs:ResultingMIMEType/text()"));
                xmldocument.setFilename( MoaXPath.getFirstString( xmldoc, "mzs:FileName/text()"));
                xmldocument.setXpath( MoaXPath.getFirstString( xmldoc, "mzs:SignatureXPath/text()"));
                String idx = MoaXPath.getFirstString( xmldoc, "mzs:SignatureXPath/@Index");
                try {
                if (idx != null) {
                    xmldocument.setXpathIndex( Integer.parseInt( idx ));
                }
                } catch (java.lang.NumberFormatException nfe) {
                    throw new MoaException("XPath Index unparseable (Integer expected)");
                }
                xmldocument.setPstylesheetName( MoaXPath.getFirstString( xmldoc, "mzs:PreviewStylesheet/mzs:FileName/text()"));

                if (xmldocument.getMimetype()==null ) {
                    throw new MoaException("XMLDocument/MIMEType or XMLProfileID missing");
                }

                if (xmldocument.getMimetype().indexOf("/")==-1) {
                    throw new MoaException ("Illegal MIMEType");
                } 

                if (xmldocument.getResultingMimetype()==null) {
                    throw new MoaException("XMLDocument/ResultingMIMEType or XMLProfileID missing");
                }

                if (xmldocument.getResultingMimetype().indexOf("/")==-1) {
                    throw new MoaException ("Illegal MIMEType");
                } 


                if (xmldocument.getFilename()==null) {
                    throw new MoaException("XMLDocument/FileName or XMLProfileID missing");
                }
                if (xmldocument.getXpath()==null) {
                    throw new MoaException("SignatureXPath or XMLProfileID missing");
                }

                try {
                    Element stylesheet = MoaXPath.getFirstElement( xmldoc, "mzs:SignatureStylesheet/mzs:XMLContent");

                    XMLSerializer serializer = new XMLSerializer();
                    ByteArrayOutputStream bos = new ByteArrayOutputStream();
                    serializer.setOutputByteStream(bos);

                    DocumentFragment df = in.createDocumentFragment();
                    if (stylesheet != null) {
                        Node tmp = stylesheet.getFirstChild();
                        while (tmp!=null && tmp.getNodeName().charAt(0)=='#') {
                            tmp = tmp.getNextSibling();
                        }
                        if (tmp!=null) {
                            df.appendChild(tmp);
                            serializer.serialize(df);
                            xmldocument.setSstylesheet( bos.toByteArray());
                        }
                    }
                } catch (java.io.IOException ioe) {
                    logger.warn( "I/O Exception", ioe);
                }

                if (xmldocument.getSstylesheet()==null) {
                    throw new MoaException("SignatureStylesheet or XMLProfileID missing");
                }
            
                try {
                    Element stylesheet = MoaXPath.getFirstElement( xmldoc, "mzs:PreviewStylesheet/mzs:XMLContent");

                    XMLSerializer serializer = new XMLSerializer();
                    ByteArrayOutputStream bos = new ByteArrayOutputStream();
                    serializer.setOutputByteStream(bos);

                    DocumentFragment df = in.createDocumentFragment();
                    if (stylesheet != null) {
                        Node tmp = stylesheet.getFirstChild();
                        while (tmp!=null && tmp.getNodeName().charAt(0)=='#') {
                            tmp = tmp.getNextSibling();
                        }
                        if (tmp!=null) {
                            df.appendChild(tmp);
                            serializer.serialize(df);
                            xmldocument.setPstylesheet( bos.toByteArray());
                        }
                    }
                } catch (java.io.IOException ioe) {
                    logger.warn( "IO Exception", ioe);
                }

 
                if (xmldocument.getPstylesheet()==null && xmldocument.getPstylesheetName()!=null && 
                    !xmldocument.getPstylesheetName().equals("") ||
                    xmldocument.getPstylesheet()!=null && (xmldocument.getPstylesheetName()==null || 
                    xmldocument.getPstylesheetName().equals(""))) {
                    throw new MoaException("PreviewStylesheet incomplete");
                }

            } else {
                MoaConfig mcfg = new MoaConfig();
                Properties xp = mcfg.getProperties("XmlProfile/"+xmldocument.getXmlProfileId());
                if (xp.size()==0)  {
                    throw new MoaException("XmlProfileID ("+xmldocument.getXmlProfileId()+") not found");
                }
            }
            try {
                Element content = MoaXPath.getFirstElement( xmldoc, "mzs:XMLContent");

                XMLSerializer serializer = new XMLSerializer();
                ByteArrayOutputStream bos = new ByteArrayOutputStream();
                serializer.setOutputByteStream(bos);

                DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                DocumentBuilder db = dbf.newDocumentBuilder();
                Document df = db.newDocument();

                if (content == null) {
                    throw new MoaException("no payload found");
                }

                Node tmp = content.getFirstChild();
                while (tmp!=null && tmp.getNodeName().charAt(0)=='#') {
                    tmp = tmp.getNextSibling();
                }
                if (tmp==null) {
                    throw new MoaException("no payload found");
                }
                df.appendChild( df.importNode(tmp, true));
               
                serializer.serialize(df);
                xmldocument.setData( bos.toByteArray());
                sess.save(xmldocument);
                r.setXmlDocument(xmldocument);

            } catch (HibernateException he) {
                logger.warn( "Hibernate Exception", he);
            } catch (java.io.IOException ioe) {
                logger.warn( "IO Exception", ioe);
            } catch (javax.xml.parsers.ParserConfigurationException pce) {
                logger.warn( "ParserConfigurationException", pce);
            }
            

        }

        Element[] base64 = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Payload/mzs:BinaryDocument[not(@xsi:nil='true')]");
        for (int i =0; i<base64.length; i++) {
            try {
                Payload p = new Payload();
                p.setRequest(r);

                //sess.disconnect();		
                String content = MoaXPath.getFirstString( base64[i], "mzs:Base64Content/text()");
                if (content!=null) {
                    p.setData( content.getBytes());
                } else {
                    throw new MoaException("no payload found");
                } 
		
                //sess.reconnect();

                p.setMimetype( MoaXPath.getFirstString( base64[i], "mzs:MIMEType/text()"));
                if (p.getMimetype()==null) throw new MoaException("no mimetype found");

                if (p.getMimetype().indexOf("/")==-1) {
                    throw new MoaException ("Illegal MIMEType");
                } 
                                    
                p.setFilename( MoaXPath.getFirstString( base64[i], "mzs:FileName/text()"));
                if (p.getFilename()==null) throw new MoaException("no filename found");
                
                p.setBase64(true);
                sess.save(p);
                r.addPayload(p);
            } catch (HibernateException he) {
                logger.warn( "Hibernate Exception", he);
            }
        }

        Element[] docref = MoaXPath.getAllElement(in.getDocumentElement(), "mzs:Payload/mzs:DocumentReference[not(@xsi:nil='true')]");
        for (int i =0; i<docref.length; i++) {
            try {
                Payload p = new Payload();
                p.setRequest(r);
                p.setUrl( MoaXPath.getFirstString( docref[i], "mzs:URL/text()"));

                if (p.getUrl()==null) throw new MoaException("no url found");
                
                p.setMimetype( MoaXPath.getFirstString( docref[i], "mzs:MIMEType/text()")); 
                if (p.getMimetype()==null) throw new MoaException("no mimetype found");
                if (p.getMimetype().indexOf("/")==-1) {
                    throw new MoaException ("Illegal MIMEType");
                } 

                p.setFilename( MoaXPath.getFirstString( docref[i], "mzs:FileName/text()"));
                if (p.getFilename()==null) throw new MoaException("no filename found");

                p.setMd5checksum( MoaXPath.getFirstString( docref[i], "mzs:MD5Checksum/text()"));
                p.setBase64(false);
                sess.save(p);
                r.addPayload(p);
            } catch (HibernateException he) {
                logger.warn( "Hibernate Exception", he);
            }

        }

        try {
            MessageContext mc = MessageContext.getCurrentContext();

            if (mc != null) {
                Message msg = mc.getRequestMessage();
                Attachments att = msg.getAttachmentsImpl();
                if (att != null) {
                    for (Iterator i = att.getAttachments().iterator(); i.hasNext(); ) {
                        try {
                            AttachmentPart ap = (AttachmentPart)i.next();
                            InputStream is = ap.getDataHandler().getInputStream();
                            ByteArrayOutputStream bos = new ByteArrayOutputStream();
                            byte[] b = new byte[1024];
                            int n = 0;
                            while ((n=is.read(b,0,1024))>0) {
                                bos.write(b,0,n);
                            }

                            Payload p = new Payload();
                            p.setRequest(r);
                            p.setData( bos.toByteArray());
                            p.setMimetype( ap.getContentType());
                            p.setBase64(false);
                            sess.save(p);
                            r.addPayload(p);
                        } catch (HibernateException he) {
                            logger.warn( "Hibernate Exception", he);
                        } catch (java.io.IOException ioe) {
                            logger.warn( "IO Exception", ioe);
                        }
                    }
                }
            }
        } catch (org.apache.axis.AxisFault af) {
            logger.warn( "Axis Fault", af);
        } catch (javax.xml.soap.SOAPException se) {
            logger.warn( "SoapException", se);
        }
        
        if (MoaXPath.getFirstElement(in.getDocumentElement(), "//mzs:Payload[not(mzs:BinaryDocument) and not(mzs:DocumentReference) or mzs:BinaryDocument[@xsi:nil='true'] and mzs:DocumentReference[@xsi:nil='true'] ]")!=null) {
            throw new MoaException("Empty Payload-Tag found");

        }
        
        if (r.getXmlDocument() == null && (r.getPayload() == null || r.getPayload().size() == 0)) {
            logger.warn( "NoPayload");
            throw new MoaException("No Payload found");
        }
        
    }
    
    /*
     * Parst Zustellserver informationen ohne zustellkopf abfrage
     */
    private static Zuseserver getZuseServer(Document in, Rcpt rcpt, boolean requiresEncryption) throws MoaException {
      // check for zuse element
      try {
        Element zuseElement = (Element) XPathAPI.selectSingleNode(in.getDocumentElement(), "mzs:Server", Constants.NAMESPACE_NODE);
        if (zuseElement == null) {
          return null;
        }
        logger.debug("Found explicit zuse server in request.");
        // check for bPK
        if (rcpt.getZusebpk() == null || rcpt.getZusebpk().length() == 0) {
          logger.debug("No bPK found in request. Aborting.");
          throw new MoaException(Constants.ERROR_REQUEST_ZBPK_MISSING, "Delivery bPK must be set when skipping delivery head query.");
        }

        Zuseserver ret = new Zuseserver();
        ret.setBpk(rcpt.getZusebpk());
        
        // parse zuseserver
        Element zuseURLElement = (Element) XPathAPI.selectSingleNode(zuseElement, "mzs:ZUSEUrlID", Constants.NAMESPACE_NODE);
        String zuseURL = DOMUtils.getText(zuseURLElement);
        ret.setUrl(zuseURL);

        Element x509Element = (Element) XPathAPI.selectSingleNode(zuseElement, "mzs:X509", Constants.NAMESPACE_NODE);
        if (x509Element != null) {
          logger.debug("X509 certificate present.");
          String x509String = DOMUtils.getText(x509Element);
          if (x509String != null && x509String.length() > 0) {
            ret.setCert(x509String.getBytes());
          }
        } else if (requiresEncryption) {
          throw new MoaException(Constants.ERROR_REQUEST_X509_MISSING, "X.509 certificate missing in request.");
        } else {
          logger.debug("No X509 certificate present.");
        }
        
        return ret;
      } catch (TransformerException e) {
        logger.warn(e.getMessage(), e);
        throw new MoaException(Constants.ERROR_ZKOPF_READING_RESPONSE, "Unclassified error reading delivery head response.");
      }
    }
    
    /**
     * @param in
     * @return String
     */
    public static String getSenderProfileId(Document in) throws MoaException {
        if (in == null) {
          return "";
        }
        String profileID = MoaXPath.getFirstString( in.getDocumentElement(), "mzs:Sender/mzs:ProfileID/text()");
        Properties prob = mcfg.getSenderInfo(profileID);
        if (prob == null || prob.isEmpty()) {
          throw new MoaException(Constants.ERROR_REQUEST_SENDER_PROFILE_NOT_FOUND, "No sender profile with name '" + profileID + "' defined.");
        }
        return profileID;
    }

    /**
     * @param in
     * @return String
     */
    public static String getSignatureProfileId(Document in) {
        if (in == null) {
          return "";
        }
        return MoaXPath.getFirstString( in.getDocumentElement(), "mzs:Sender/mzs:SignatureKeyID/text()");
    }

    /**
     * @param in
     * @return String
     */
    public static String getAppDeliveryId(Document in) {
        if (in == null) {
          return "";
        }
        return MoaXPath.getFirstString( in.getDocumentElement(), "mzs:MetaData/mzs:AppDeliveryID/text()");
    }
    
    /**
     * @param in
     * @return boolean
     */
    public static boolean getPrivateDelivery(Document in, String senderProfileId) {
        if (in == null) {
          return false;
        }
        String privateDeliveryString = MoaXPath.getFirstString( in.getDocumentElement(), "mzs:MetaData/mzs:PrivateDelivery/text()");
        
        // element not present
        if (privateDeliveryString == null) {
          return mcfg.isPrivateDeliveryOnly(senderProfileId);
        }
        
        // try to parse
        if ("true".equals(privateDeliveryString.toLowerCase())) {
          return true;
        }
        return false;
    }

    public static String getDocumentClass(Document in) throws MoaException {
      if (in == null) {
        return null;
      }
      String documentClassString = MoaXPath.getFirstString( in.getDocumentElement(), "mzs:MetaData/mzs:DocumentClass/text()");
      if (documentClassString != null) {
        if (!documentClassString.startsWith("x-edclass/")) {
          throw new MoaException(Constants.ERROR_REQUEST_DOCUMENT_CLASS_INVALID_FORMAT, "Document class must start with x-edclass");
        }
      }
      return documentClassString;
  }

    
    public static DualDelivery getDualDelivery(Document in) throws MoaException {
      if (in == null) {
        return null;
      }
      Element dualDeliveryElement = MoaXPath.getFirstElement(in.getDocumentElement(), "mzs:MetaData/mzs:DualDelivery");

      if (dualDeliveryElement == null) {
        return null;
      }
      // DualDelivery element present, return instance
      DualDelivery dualDelivery = new DualDelivery();
      
      // check for dual delivery server
      String dualDeliveryServerString = MoaXPath.getFirstString( in.getDocumentElement(), "mzs:MetaData/mzs:DualDelivery/mzs:DualDeliveryServer/text()");
      dualDelivery.setDualDeliveryServer(dualDeliveryServerString);
      
      // print parameters
      Element printParametersElement = MoaXPath.getFirstElement(in.getDocumentElement(), "mzs:MetaData/mzs:DualDelivery/mzs:PrintParameter");
      if (printParametersElement != null) {
        try {
          DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
          factory.setNamespaceAware(true);
          DocumentBuilder builder = factory.newDocumentBuilder();
          Document ppDoc = builder.newDocument();
          ppDoc.appendChild(ppDoc.importNode(printParametersElement, true));
          
          if (printParametersElement.getPrefix() != null) {
            ppDoc.getDocumentElement().setAttribute("xmlns:" + printParametersElement.getPrefix(), printParametersElement.getNamespaceURI());
          } else {
            ppDoc.getDocumentElement().setAttribute("xmlns", printParametersElement.getNamespaceURI());
          }
          
          dualDelivery.setPrintParameter(MoaXMLUtils.serializeToByteArray(ppDoc, true));
        } catch (IOException e) {
          throw new MoaException(e);
        } catch (ParserConfigurationException e) {
          throw new MoaException(e);
        }
      }
      
      return dualDelivery;
      
  }
    
    public static boolean getSync(Document in) {
      if (in == null) {
        return false;
      }
      String syncString = in.getDocumentElement().getAttribute("sync");
      
      // element not present
      if (syncString == null) {
        return false;
      }
      // try to parse
      if ("true".equals(syncString.toLowerCase())) {
        return true;
      }
      return false;
  }


    /**
     * @param in
     * @return String
     * @throws MoaException
     */
    public static String getDeliveryQuality(Document in) throws MoaException {
        if (in == null) {
          return "";
        }
        String test = MoaXPath.getFirstString( in.getDocumentElement(), "mzs:MetaData/mzs:DeliveryQuality/text()");
        if (test == null) throw new MoaException("DeliveryQuality missing");
        return test;
    }

    /**
     * @param in
     * @return boolean
     * @throws MoaException
     */
    public static boolean getRequiresEncryption(Document in) throws MoaException {
        if (in == null) return false;

        String test = MoaXPath.getFirstString( in.getDocumentElement(), "mzs:MetaData/mzs:RequiresEncryption/text()");
        if (test == null) {
          throw new MoaException("RequiresEncryption missing");
        }
        if (test.equals("true") || test.equals("false")) {
            return test.equals("true");
        } else {
            throw new MoaException("RequiresEncryption must be 'true' or 'false'");
        }
    }

    /**
     * @param rcpt
     * @return boolean
     */
    public static boolean hasBPK(Rcpt rcpt) {
        Set ids = rcpt.getIdentification();
        if (ids == null || ids.size()==0) return false;
        return true;
    }

    /**
     * @param rcpt
     * @return boolean
     */
    public static boolean hasPostAddress( Rcpt rcpt) {
        Set post = rcpt.getPostaladdress();
        if (post == null || post.size()==0) return false;
        return true;
   
    }

    /**
     * @param rcpt
     * @return boolean
     */
    public static boolean hasInternetAddress( Rcpt rcpt) {
        Set inet = rcpt.getInternetaddress();
        if (inet == null || inet.size()==0) return false;
        return true;
    }

    /**
     * @param payload
     * @return boolean
     */
    public static boolean isPartialSuccess( Set payload ) {
        if (payload != null) {
            for (Iterator i = payload.iterator(); i.hasNext(); ) {
            Payload p = (Payload)i.next();
                if (p.getUrl() != null && !p.getUrl().equals("")) {
                    return true;
                }                                    
            } 
        }
        return false;
    }     
}
