/*
 * Copyright 2004 Bundeskanzleramt sterreich
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package moazs.api;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import moazs.util.MoaConfig;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
* Hilfklasse zum erzeugen der Antworten im Synchronen Teil von MoaZS
*/
public class MoaResponse {
  
    private static String NS_MZS;
    //private static String NS_PERS;
    static private Logger logger = Logger.getLogger(MoaResponse.class);

    static {
        MoaConfig mcfg = new MoaConfig();
        NS_MZS = mcfg.getProperty("mzs", "Namespaces");
        //NS_PERS= mcfg.getProperty("p1",  "Namespaces");
    }

    /**
    * Antwort im Erfolgsfall
     * @param appid AppDeliveryID
     * @param mzsid MZSDeliveryID
     * @return DOM-Document
     */
    public static Document getSuccess(String appid, String mzsid) {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document result = db.newDocument();
            Element root = result.createElementNS(NS_MZS, "DeliveryResponse");
            Element success = result.createElementNS(NS_MZS, "Success");
            Element app_id  = result.createElementNS(NS_MZS, "AppDeliveryID");
            app_id.appendChild( result.createTextNode(appid));
            Element mzs_id  = result.createElementNS(NS_MZS, "MZSDeliveryID");
            mzs_id.appendChild( result.createTextNode(mzsid));
            result.appendChild(root);
            root.appendChild(success);
            success.appendChild(app_id);
            success.appendChild(mzs_id);

            return result;
        } catch (javax.xml.parsers.ParserConfigurationException pce) {
            logger.warn( "ParserConfigException", pce);
            return null;
        }

    }

    /**
    * Antwort bei bergabe eines HttpCallbacks
     * @param appid AppDeliveryID
     * @param mzsid MZSDeliveryID
     * @return DOM-Document
     */
    public static Document getPartialSuccess(String appid, String mzsid) {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document result = db.newDocument();
            Element root = result.createElementNS(NS_MZS, "DeliveryResponse");
            Element success = result.createElementNS(NS_MZS, "PartialSuccess");
            Element app_id  = result.createElementNS(NS_MZS, "AppDeliveryID");
            app_id.appendChild( result.createTextNode(appid));
            Element mzs_id  = result.createElementNS(NS_MZS, "MZSDeliveryID");
            mzs_id.appendChild( result.createTextNode(mzsid));
            result.appendChild(root);
            root.appendChild(success);
            success.appendChild(app_id);
            success.appendChild(mzs_id);

            return result;
        } catch (javax.xml.parsers.ParserConfigurationException pce) {
            logger.warn( "ParserConfigException", pce);
            return null;
        }

    }
    
    /**
    * Antwort im Fehlerfall
     * @param appid AppDeliveryID
     * @param mzsid MZSDeliveryID
     * @param code ErrorCode
     * @param text ErrorText
     * @param url Document-Referenz
     * @return DOM-Document
     */
    public static Document getError(String appid, String mzsid, String code, String text, String url) {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document result = db.newDocument();
            Element root = result.createElementNS(NS_MZS, "DeliveryResponse");
            Element error = result.createElementNS(NS_MZS, "Error");
            Element app_id  = result.createElementNS(NS_MZS, "AppDeliveryID");
            app_id.appendChild( result.createTextNode(appid));
            Element mzs_id  = result.createElementNS(NS_MZS, "MZSDeliveryID");
            mzs_id.appendChild( result.createTextNode(mzsid));
            Element ecode  = result.createElementNS(NS_MZS, "Code");
            ecode.appendChild( result.createTextNode(code));
            Element etext  = result.createElementNS(NS_MZS, "Text");
            etext.appendChild( result.createTextNode(text));

            result.appendChild(root);
            root.appendChild(error);
            error.appendChild(app_id);
            error.appendChild(mzs_id);
            error.appendChild(ecode);
            error.appendChild(etext);

            if (url != null) {
                Element eurl  = result.createElementNS(NS_MZS, "DocumentReference");
                eurl.appendChild( result.createTextNode(url));
                error.appendChild(eurl);
            }

            return result;
        } catch (javax.xml.parsers.ParserConfigurationException pce) {
            logger.warn( "ParserConfigException", pce);
            return null;
        }
    }


}
