/*
 * Copyright 2004-2009 Bundeskanzleramt sterreich
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package moazs.async;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.security.MessageDigest;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import javax.mail.MessagingException;

import moazs.external.AppNotification;
import moazs.external.MoaDelivery;
import moazs.external.MoaSMIME;
import moazs.external.MoaSS;
import moazs.modell.Payload;
import moazs.modell.Request;
import moazs.modell.Status;
import moazs.modell.XmlDocument;
import moazs.util.HibernateUtil;
import moazs.util.MoaConfig;
import moazs.util.MoaDeckblatt;
import moazs.util.MoaFactory;
import net.sf.workqueue.QueueException;
import net.sf.workqueue.WorkerThread;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.log4j.Logger;
import org.bouncycastle.util.encoders.Hex;
import org.hibernate.HibernateException;
import org.hibernate.ObjectNotFoundException;
import org.hibernate.Transaction;
import org.hibernate.classic.Session;

/**
* Workerthread der Die Requests im Asyncronen Teil abarbeitet.
*/
public class MoaWorker extends WorkerThread {

    static private Logger logger = Logger.getLogger(MoaWorker.class);
    
    private MoaConfig config = null;
    
    public MoaWorker() {
      config = new MoaConfig();
    }
    
    /**
    * Methode die von der Mainloop der WorkerThread klasse aufgerufen wird
    * wenn aus der Queue erfolgreich eine ID gelesen werden konnte.
    * Verteilt die Requests je nach status auf die einzelnen methoden.
    * wenn der Returnwert der Methoden &gt; 0 ist wird der wert als neuer 
    * status verwendet und der Request wieder in die queue eingetragen.
    * Falls der Returnwert  &lt; 0 ist wird der Request mit der QueueID -1 
    * gespeichert, oder er ist bereits gelscht worden und wird ignoriert.
    */
    public void doSomething(Object o) {
        logger.debug("Async: doSomething ...");
        logger.debug("REQUESTCOUNT " + moazs.api.MoaSync.getCount());
        if (moazs.api.MoaSync.getCount() > 5) {
            try {
                int w = moazs.api.MoaSync.getCount() * 5000; 
                logger.debug("Schlafe: " + w);
                Thread.sleep(w < 30000 ? w : 30000);
            } catch (Exception e) {
              logger.warn( "Schwerer Fehler: " + e.getMessage());
            }
        }
        int result = -1;
        Request req = null;
        boolean error = true;
        Session sess = HibernateUtil.getCurrentSession();
        
        // create transaction
        Transaction tx = sess.beginTransaction();
        
        try {
            req = (Request)sess.get(Request.class, (Long)o);
            if (req == null) {
              throw new ObjectNotFoundException((Long)o, Request.class.toString());
            }
            switch ((int)req.getStatus().getId()) {
                case 0:
                    result = newRequest(sess, req);
                    break;
                case 1:
                    result = signRequest(sess, req);
                    break;
                case 2:
                    result = mimeRequest(sess, req);
                    break;
                case 3:
                    result = sendRequest(sess, req);
                    break;
                case 4:
                    result = deleteRequest(sess, req);
                    break;
            }

            if (result >= 0) { 
                req.setStatus( (Status)sess.load(Status.class, new Long( result )));
                if (moazs.api.MoaSync.getCount()<5) {
                    req.setQueueId( MoaFactory.getQueueId());
                } else {
                    req.setQueueId( -1 );
                }
            } 
            tx.commit();
            error = false; 
        } catch (HibernateException he) {
            logger.warn( "Hibernate Exeption ...", he);
        } catch (Exception e) {
            logger.warn( "General Exception ...", e);
        } catch (Throwable e) {
            logger.warn( "General Throwable ...", e);
        } finally {
          try {
            if (error) {
              logger.debug("rollback ...");
              try {
                if (HibernateUtil.getCurrentSession().getTransaction().isActive()) {
                  HibernateUtil.getCurrentSession().getTransaction().rollback();
                }
              } catch(HibernateException err) {
                logger.warn( "Could not rollback transaction.");
              }
            }
            //sess.close(); 
            if (!error && result >= 0) {
              logger.debug("worked ...");
              if ( req.getQueueId()>=0 ) {
                reenqueue(req.getId());
              }
            }
          } catch (HibernateException he) {
            logger.warn( "Hibernate Exeption ...", he);
          }
        }
    }

    /**
    * Methode die einen neuen Request aufbereitet und versucht HttpCallback urls aufzulsen 
    * und ein XMLDeckblatt zu erstellen.
    * Falls eine XMLProfileID angegeben wurde werden die Daten aus der Konfiguration in
    * das XMLDocument-Objekt eingetragen.
    * Wird im Fehlerfall wiederholt.
    */
    private int newRequest(Session sess, Request r) {
        logger.debug("Async: new ...");

        MoaConfig mcfg  = new MoaConfig();
        Properties info = mcfg.getSenderInfo(r.getSenderProfileId());
        boolean skipDeckblatt = false;
        //if ("yes".equals((String)info.getProperty("SkipDeckblatt")) && (r.getSignatureProfileId() == null || r.getSignatureProfileId().equals(""))) {
        // check if we have to skip deckblatt property
        String skipDeckblattProperty = info.getProperty("SkipDeckblatt", "false"); 
        if ("yes".equals(skipDeckblattProperty) ||
            "true".equals(skipDeckblattProperty)) {
            skipDeckblatt = true;
        }
        /* XMLDeckblatt
         * 
         * Will be generated only if SkipDeckblatt Parameter = false AND only binary payload present
         */
        if (!skipDeckblatt && r.getXmlDocument() == null) {
            try {
                MoaDeckblatt dm = new MoaDeckblatt();
                XmlDocument xml = dm.makeDeckblatt(r);
                sess.save(xml);
                r.setXmlDocument(xml);
            } catch (HibernateException he) {
                logger.warn( "HibernateException ...", he);
            }
        } else {
            logger.info( "Skipping 'deckblatt' generation ...");
        }
        
        /* XMlProfile */
        if ( r.getXmlDocument()!=null && r.getXmlDocument().getXmlProfileId() != null) {
            logger.info("Async: XmlProfile nachladen ...");

            XmlDocument xdoc = r.getXmlDocument();
            Properties xp = mcfg.getProperties("XmlProfile/"+xdoc.getXmlProfileId());
            xdoc.setFilename( xp.getProperty("FileName"));
            xdoc.setMimetype( xp.getProperty("MIMEType"));
            xdoc.setResultingMimetype( xp.getProperty("ResultingMIMEType"));
            xdoc.setXpath( xp.getProperty("SignatureXPath"));
            xdoc.setXpathIndex( Integer.parseInt( xp.getProperty("SignatureXPathIndex")));
            xdoc.setPstylesheetName( xp.getProperty("PreviewStylesheetName"));
            /* Signatur Stylesheet */
            try {
                if (xp.getProperty("SignatureStylesheet")!=null && !xp.getProperty("SignatureStylesheet").equals("")) {
                    BufferedReader in = new BufferedReader(new FileReader( xp.getProperty("SignatureStylesheet") ));
                    StringBuffer buffer = new StringBuffer();
                    String dummy = in.readLine();
                    while (dummy != null) {
                        buffer.append(dummy);
                        dummy = in.readLine();
                    }
                    xdoc.setSstylesheet( buffer.toString().getBytes());
                } else {
                  xdoc.setSstylesheet((byte[]) null);
                }
                
            } catch (java.io.FileNotFoundException fnf) {
                logger.warn( "File Not Found", fnf);
            } catch (java.io.IOException ioe) {
                logger.warn( "IOException", ioe);
            } 
            /* PreviewStylesheet */
            try { 
                BufferedReader in = new BufferedReader(new FileReader( xp.getProperty("PreviewStylesheet") ));
                StringBuffer buffer = new StringBuffer();
                String dummy = in.readLine();
                while (dummy != null) {
                    buffer.append(dummy);
                    dummy = in.readLine();
                }
                xdoc.setPstylesheet( buffer.toString().getBytes());
            } catch (java.io.FileNotFoundException fnf) {
                logger.warn( "File Not Found", fnf);
            } catch (java.io.IOException ioe) {
                logger.warn( "IOException", ioe);
            } 
        }

        /* Payload Aufbereiten */
        Set payload  = r.getPayload();
        boolean error = false;
        String brokenURL = "";
        
        if (payload != null) {
            for (Iterator i = payload.iterator(); i.hasNext(); ) {
                Payload p = (Payload)i.next();
                if (p.getUrl() != null && !p.getUrl().equals("")) {
                    try {
                        HttpClient client = new HttpClient();                    
                        logger.info ( p.getUrl());               
                        GetMethod gm = new GetMethod(p.getUrl());
                        client.executeMethod(gm);
                        p.setData( gm.getResponseBody());
                        logger.debug(new String(p.getDataAsByteArray()));
                        if (p.getMd5checksum()!=null && !p.getMd5checksum().equals("") ) {
                            logger.debug("Checking MD5");
                            try {
                                MessageDigest md5 = MessageDigest.getInstance("MD5");
                                md5.update(p.getDataAsByteArray());
                                byte[] digest = md5.digest();
                                logger.debug(new String( Hex.encode(digest))); 
                                logger.debug(p.getMd5checksum());

                                if (!p.getMd5checksum().equals( new String( Hex.encode(digest)))) {
                                    AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
                                    appn.sendError("MZS"+r.getId(), "500", "MD5 Prfsumme falsch", brokenURL);
                                    return 4;

                                }
                            } catch ( java.security.NoSuchAlgorithmException nsae) {
                                AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
                                appn.sendError("MZS" + r.getId(), "500", "MD5 kann nicht geprft werden", brokenURL);
                                return 4;

                            }
                        }
                       
                    } catch ( java.io.IOException ioe) {
                        logger.warn( "IO Exeption ...", ioe);
                        error = true;
                        brokenURL = p.getUrl();
                    }
                }
            }
        }

        /* Im Fehlerfall neu versuchen */
        if (error) {
            if (r.getRetryCount() < Integer.parseInt(mcfg.getProperty("MaxRetry","clientapp"))) {
                logger.warn( "Retrying APPID="+r.getAppDeliveryId() + " (" + r.getRetryCount() + ")"); 
                r.setRetryCount(r.getRetryCount() +1);
                r.setRetryTs( new Date( new Date().getTime() + Integer.parseInt(mcfg.getProperty("RetryTimeout", "clientapp")) * 1000));
                r.setQueueId(-1);
                return -1;
            } else {

                AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
                appn.sendError("MZS" + r.getId(), "500", "Kann Attachments nicht laden", brokenURL);
                return 4;
            }
        }
        
        r.setRetryCount(0);
        logger.debug("return ...");
        return 1;
    }

    /**
    * Methode mit der die ServerSignatur erstellt wird. 
    * Wird im Fehlerfall wiederholt.
    */
    private int signRequest(Session sess, Request r) {
        boolean error = false;
        try {
            logger.info("Async: sign ...");
            if (r.getSignatureProfileId() == null || r.getSignatureProfileId().equals("")) {
                logger.info("Async: Skipping signature ...");            
                r.setRetryCount(0);
                return 2;
            } 

            MoaSS moass = new MoaSS();
	    
            XmlDocument xml = r.getXmlDocument();
            byte[] data = moass.sign(xml, r.getPayload(), r.getSignatureProfileId(), sess);
            if (data == null) {
                error = true;
                throw new moazs.api.MoaException("Signature failed ...");
            }
	    
            xml.setData(data);
            r.setRetryCount(0);
            return 2;

        } catch (moazs.api.MoaException me) {
            logger.warn( "MoaException ...", me);
        }
        
        MoaConfig mcfg = new MoaConfig();
        if (!error && r.getRetryCount() < Integer.parseInt(mcfg.getProperty("MaxRetry","moass"))) {
            logger.warn( "Retrying APPID="+r.getAppDeliveryId()); 
            r.setRetryCount(r.getRetryCount() +1);
            r.setRetryTs( new Date( new Date().getTime() + Integer.parseInt(mcfg.getProperty("RetryTimeout", "moass")) * 1000));
            r.setQueueId(-1);
            return -1;
        }
        
        AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
        appn.sendError("MZS"+r.getId(), "500", "Nachricht kann nicht Signiert werden", null);

        return 4;
    }

    /**
    * Methode mit der der Mime kontainer fr die Zustellung erstellt wird. 
    * Je nach konfiguration wird ein SMIME oder ein Zuse Kontainer erstellt.
    * Bei bedarf wird dieser dann Verschlsselt.
    */
    private int mimeRequest(Session sess, Request r) {
        try {
            logger.info("Async: mime/encrypt ...");
            MoaSMIME smime = new MoaSMIME();

            smime.makeMessage(r.getXmlDocument(), r.getPayload());
            boolean crypt = false;
            if (r.getZuseserver() != null &&
                r.getZuseserver().getCertAsByteArray() != null &&
                r.getZuseserver().getCertAsByteArray().length != 0) {
                logger.debug("encryption ...");
                crypt = true;
                byte[] cert = r.getZuseserver().getCertAsByteArray();
                try {
                    //sess.flush();
                    //sess.connection().commit();
                    //sess.disconnect();
                    smime.encrypt(cert);
                    //sess.reconnect();
                /*} catch (java.sql.SQLException se) {
                    throw new moazs.api.MoaException("Fehler bei commit ...");*/
                } catch (moazs.api.MoaException me) {
                    //sess.reconnect();
                    throw me;
                }
            }
            Payload p = new Payload();
            logger.debug("clear old payload ...");
            byte[] data = smime.getResult();
            if (r.getPayload()!=null) {
                for (Iterator i = r.getPayload().iterator(); i.hasNext(); ) {
                    sess.delete( (Payload)i.next());
                }
                r.getPayload().clear();
            }
            logger.debug("add payload ...");

            p.setData( data );
            p.setRequest(r);
            p.setMimetype(crypt?"application/pkcs7-mime":"multipart/mixed");
            sess.save(p);
            r.addPayload(p);
            //sess.flush();
            logger.debug("reenqueue request ...");

            r.setRetryCount(0);
            return 3;
        } catch (HibernateException he) {
            logger.warn( "HibernateExeption .("+r.getId()+")..", he);
            logger.warn( "APPID="+r.getAppDeliveryId()); 

            AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
            appn.sendError("MZS"+r.getId(), "500", "Nachricht kann nicht verschlsselt werden", null);

            return 4;

        } catch (moazs.api.MoaException me) {
            logger.warn( "MOAException ...", me);
            logger.warn( "APPID="+r.getAppDeliveryId()); 
            AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
            appn.sendError("MZS"+r.getId(), "500", "Nachricht kann nicht verschlsselt werden", null);

            return 4;
        } catch (IOException he) {
          logger.warn( "I/O Exception .("+r.getId()+")..", he);
          logger.warn( "APPID="+r.getAppDeliveryId()); 
          AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
          appn.sendError("MZS"+r.getId(), "500", "Nachricht kann nicht verschlsselt werden", null);
          return 4;
        } catch (MessagingException e) {
          logger.warn( "Messaging exception .("+r.getId()+")..", e);
          logger.warn( "APPID="+r.getAppDeliveryId()); 
          AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
          appn.sendError("MZS"+r.getId(), "500", "Nachricht kann nicht verschlsselt werden", null);
          return 4;
        }
    }

    /**
    * Methode mit der das Zustellstck an den ZuseServer bergeben wird.
    * Wird im Fehlerfall wiederholt.
    */
    private int sendRequest(Session sess, Request r) {
        logger.info("Async: send ...");
        
        
        
        int status = new MoaDelivery().createDeliveryRequest(sess, r);
        if (status < 0) {
            MoaConfig mcfg = new MoaConfig();
            if (r.getRetryCount() < Integer.parseInt(mcfg.getProperty("MaxRetry","MoaDelivery"))) {
                logger.warn( "Retrying APPID="+r.getAppDeliveryId() + " (" + r.getRetryCount() + ")"); 
                r.setRetryCount(r.getRetryCount() + 1);
                r.setRetryTs( new Date( new Date().getTime() + Integer.parseInt(mcfg.getProperty("RetryTimeout", "MoaDelivery")) * 1000));
                r.setQueueId(-1);
                return -1;
            }
        }

        /* Was auch immer passiert - nach diesem Punkt wird der Request gelscht */
        /* Falls status < 0 (d.h. exception beim zustellen - z.b. keine verbindung o., sende notification) */
        if (status < 0) {
          AppNotification appn = new AppNotification(r.getSenderProfileId(), r.getAppDeliveryId(), r.getSyncId());
          appn.sendError(new String().valueOf(r.getId()), "551", "Allgemeiner Fehler beim Zustellen", "referenz");
        }
        
        r.setRetryCount(0);
        return 4;
    }

    /**
    * Methode mit der die Zustellstcke nach bergabe an den Zustellserver
    * oder im Fehlerfall gelscht wird.
    */
    private int deleteRequest(Session sess, Request r) {
        try {
            logger.info("Async: delete ("+r.getId()+")...");
            sess.delete(r);
            //sess.flush();

        } catch (HibernateException he) {
            logger.warn( "HibernateExeption ...", he);
        }
        return -1;
    }

    /**
    * Methode die den Request wieder in die Queue einsortiert.
    */
    private void reenqueue(long id) {
        logger.debug("reenqueue ...");
        try {
            getQueue().add( new Long(id) );
        } catch (QueueException qe) {
            Session sess = HibernateUtil.getCurrentSession();
            Transaction tx = sess.beginTransaction();
            boolean error = true;
            try {
                Request req = (Request)sess.get(Request.class, new Long(id));
                if (req == null) {
                  throw new ObjectNotFoundException(new Long(id), Request.class.toString());
                }
                req.setQueueId(-1);
                //sess.connection().commit();
                tx.commit();
                error = false;
            } catch (HibernateException he) {
                logger.warn( "Hibernate Exeption ...", he);
            } catch (Throwable he) {
                logger.warn( "Throwable ...", he);
            } finally {
              if (error) {
                logger.debug("rollback ...");
                try {
                  if (HibernateUtil.getCurrentSession().getTransaction().isActive()) {
                    HibernateUtil.getCurrentSession().getTransaction().rollback();
                  }
                } catch(HibernateException err) {
                  logger.warn( "Could not rollback transaction.");
                }
              }
            }

        }
    }
}

