package moazs.szr;

import java.io.File;
import java.net.URL;
import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;

import moazs.util.MoaConfig;

import org.apache.axis.AxisFault;
import org.apache.log4j.Logger;

import at.gv.bmi.szr.FremdBPKRequestType;
import at.gv.bmi.szr.FremdBPKType;
import at.gv.bmi.szr.IdentityLinkType;
import at.gv.bmi.szr.PersonInfoType;
import at.gv.bmi.szr.SZR;
import at.gv.bmi.szr.SZRServiceLocator;
import at.gv.bmi.szr.SZRSoapBindingStub;
import at.gv.bmi.szr.client.PVPHeader;
import at.gv.bmi.szr.key.KeyValueType;
import at.gv.bmi.szr.person.PersonNameType;
import at.gv.bmi.szr.person.PhysicalPersonType;

public class SZRClient {

  private SZR stub;

  private SZRServiceLocator locator;

  private final float PVP_STANDARD_VERSION = 1.8f;
  private final String IDENTIFIER_ORGANWALTER = "OW";
  
  static private Logger logger = Logger.getLogger(SZRClient.class);

  private MoaConfig cfg = null;
  
  public SZRClient(URL url, String pvpFilePath, float pvpVersion)
      throws SZRClientException {
    cfg = new MoaConfig();
    /*try {
      Provider provider = (Provider) Class.forName(
          "com.sun.net.ssl.internal.ssl.Provider").newInstance();
      Security.addProvider(provider);
      System.setProperty("java.protocol.handler.pkgs",
          "com.sun.net.ssl.internal.www.protocol");
    } catch (Exception e) {
      throw new SZRClientException("Can't initialize the ssl provider. "
          + e.getMessage());
    }*/

    /*System.setProperty("javax.net.ssl.keyStore", keyStore);
    System.setProperty("javax.net.ssl.keyStorePassword", keyStorePw);
    if (keyStoreType != null) {
      System.setProperty("javax.net.ssl.keyStoreType", keyStoreType);
    }
    System.setProperty("javax.net.ssl.trustStore", trustStore);
    System.setProperty("javax.net.ssl.trustStorePassword", trustStorePw);
    if (trustStoreType != null) {
      System.setProperty("javax.net.ssl.trustStoreType", trustStoreType);
    }*/
    
    File pvpFile = new File(pvpFilePath);

    try {
      initConnection(url, pvpFile, pvpVersion);
    } catch (ServiceException e) {
      throw new SZRClientException("Can't initialize the connection. "
          + e.getMessage());
    }

  }

  public void initConnection(URL url, File pvpFile, float pvpVersion)
      throws ServiceException {
    locator = new SZRServiceLocator();
    // new client
    /*if (isTestEnvironment) {
      stub = this.locator.getSZRTestumgebung(url);
    } else {
      stub = this.locator.getSZRProduktion(url);
    }*/
    // old client
    stub = this.locator.getSZR(url);
    // pvp informationen
    /*
     * New client
    at.gv.bmi.szr.wsse.Security security = new at.gv.bmi.szr.wsse.Security();
    security.setPvpToken(SZRGWConfiguration.getPVPToken());
    ((SZRSoapBindingStub) stub).setHeader("http://schemas.xmlsoap.org/ws/2002/04/secext", "Security", security);
    */
    
    // old client
    if (pvpFile != null) {
      if (pvpVersion ==  -1) {
        pvpVersion = PVP_STANDARD_VERSION;
      }
      try {
        PVPHeader pvpHeader = new PVPHeader(pvpVersion, pvpFile);
        ((SZRSoapBindingStub) this.stub).setPVPHeader(pvpHeader);
      } catch(Throwable t) {
        logger.warn("Fehler beim initialisieren der SZR Verbindung: ", t);
      }
    }
  }

  public String processGetStammzahl(PersonInfoType personInfoType)
      throws SZRClientException {
    try {
      return stub.getStammzahl(personInfoType);
    } catch (RemoteException e) {
      if (e instanceof AxisFault) {
        AxisFault af = (AxisFault) e;
        // parse axisfault
        String codeString = af.getFaultCode().getLocalPart();
        int error = -1;
        try {
          error = Integer.parseInt(codeString.substring(1));
        } catch(NumberFormatException err) {
        }
        
        throw new SZRClientException(af.getFaultString(), error);
      } else {
        throw new SZRClientException(e.getMessage());
      }
    }
  }
  
  public String processGetOrganwalterBPK(PersonInfoType personInfoType) throws SZRClientException {
    return processGetBPK(personInfoType, "urn:publicid:gv.at:cdid+" + IDENTIFIER_ORGANWALTER);
  }
  

  public String processGetBPK(PersonInfoType personInfoType, String bereich)
      throws SZRClientException {
    try {
      return stub.getBPK(personInfoType, bereich);
    } catch (RemoteException e) {
      e.printStackTrace();
      if (e instanceof AxisFault) {
        AxisFault af = (AxisFault) e;
        // parse axisfault
        String codeString = af.getFaultCode().getLocalPart();
        int error = -1;
        try {
          error = Integer.parseInt(codeString.substring(1));
        } catch(NumberFormatException err) {
        }
        
        throw new SZRClientException(af.getFaultString(), error);
      } else {
        throw new SZRClientException(e.getMessage());
      }
    }
  }

  public IdentityLinkType processGetIdentityLink(PersonInfoType personInfoType,
      KeyValueType[] keys) throws SZRClientException {
    try {
      return stub.getIdentityLink(personInfoType, keys);
    } catch (RemoteException e) {
      if (e instanceof AxisFault) {
        AxisFault af = (AxisFault) e;
        // parse axisfault
        String codeString = af.getFaultCode().getLocalPart();
        int error = -1;
        try {
          error = Integer.parseInt(codeString.substring(1));
        } catch(NumberFormatException err) {
        }
        
        throw new SZRClientException(af.getFaultString(), error);
      } else {
        throw new SZRClientException(e.getMessage());
      }
    }
  }

  public FremdBPKType processTranslateBPK(PersonInfoType personInfoType,
      String bPK, String inputTarget, String outputTarget, String vkz, String reason)
      throws SZRClientException {

    FremdBPKRequestType fremdType = new FremdBPKRequestType();
    fremdType.setBereichsKennung(outputTarget);
    fremdType.setVKZ(vkz);
    FremdBPKRequestType[] target = new FremdBPKRequestType[] { fremdType };

    try {
      return stub. transformBPK(personInfoType, bPK, inputTarget,
          reason, target)[0];
    } catch (RemoteException e) {
      if (e instanceof AxisFault) {
        AxisFault af = (AxisFault) e;
        // parse axisfault
        String codeString = af.getFaultCode().getLocalPart();
        int error = -1;
        try {
          error = Integer.parseInt(codeString.substring(1));
        } catch(NumberFormatException err) {
        }
        
        throw new SZRClientException(af.getFaultString(), error);
      } else {
        throw new SZRClientException(e.getMessage());
      }
    }
  }

  public PersonInfoType createPerson(String givenName, String familyName,
      String birthDay) throws SZRClientException {
      PersonNameType name = new PersonNameType();
      name.setGivenName(givenName);
      name.setFamilyName(familyName);
      PhysicalPersonType person = new PhysicalPersonType();
      person.setDateOfBirth(birthDay);
      person.setName(name);
      PersonInfoType personInfo = new PersonInfoType();
      personInfo.setPerson(person);
      return personInfo;
  }

}
